/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sockios.h>
#include <linux/proc_fs.h>
#ifdef CONFIG_FILTER_UNM
#include <linux/filter.h>
#endif

#include <linux/version.h>
#include <linux/autoconf.h>
#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
#define MODVERSIONS
#endif
#if defined(MODVERSIONS) && !defined(__GENKSYMS__)
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#include <linux/modversions.h>
#endif
#endif
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/socket.h>
#include <linux/in.h>
#include <linux/unistd.h>
#include <linux/pagemap.h>
#include <linux/pci.h>
#include <linux/net.h>
#include <linux/poll.h>
#include <linux/skbuff.h>
#include <net/protocol.h>
#include <net/udp.h>
#include <net/ip.h>
#include <linux/if.h>
#include <linux/if_arp.h>
#include <linux/inetdevice.h>
#include <net/route.h>
#include <net/tcp.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,14)
#include <net/inet_hashtables.h>
#endif

#include <linux/inetdevice.h>
#include <linux/if.h>
#include <linux/if_vlan.h>

#include <unm_inc.h>
#include <pegnet_inc.h>
#include "nic_cmn.h"
#include "hostInterface.h"
#include "flow_table.h"
#include "unm_host_pnic_types.h"

#include "driver_info.h"
#include "unm_version.h"

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(LSA_DRIVER_DESCRIPTION);
MODULE_LICENSE(LSA_DRIVER_LICENSE);
MODULE_VERSION(UNM_NIC_LINUX_VERSIONID);	/* We build all 3 modules
						   together so the same
						   version is fine. */

#define NETXEN_PCI_ID(device_id)   PCI_DEVICE(PCI_VENDOR_ID_NX, device_id)

static struct pci_device_id unm_pci_tbl[] __devinitdata = {
        { NETXEN_PCI_ID(PCI_DEVICE_ID_NX_QG) },
        { NETXEN_PCI_ID(PCI_DEVICE_ID_NX_XG) },
        { NETXEN_PCI_ID(PCI_DEVICE_ID_NX_CX4) },
        { NETXEN_PCI_ID(PCI_DEVICE_ID_NX_IMEZ) },
        { NETXEN_PCI_ID(PCI_DEVICE_ID_NX_HMEZ) },
        { NETXEN_PCI_ID(PCI_DEVICE_ID_NX_IMEZ_DUP) },
        { NETXEN_PCI_ID(PCI_DEVICE_ID_NX_HMEZ_DUP) },
        {0,}
};

MODULE_DEVICE_TABLE(pci, unm_pci_tbl);

/*
 * Function declaration of static functions.
 */
static void nx_sock_cleanup_task(void *v);
static int unm_disconnect(struct socket *sock, int flags, long timeo);

/*
 * Global variables.
 */
nx_tnic_dbg_cntrs_t	nx_tnic_dbg_cntrs;
unsigned int max_tcp_conn = UNM_MAX_TCP_CONNS;


/* Whenever possible compress qmsg IOVs before passing it to the card */
extern int			global_use_compressed_iov;
/* Size beyond which use zero copy transmits, if possible */
extern uint32_t                 tx_zero_copy_size_thresh;
/* Connection threshold beyond which use zero copy transmits, if possible */
extern uint32_t                 tx_zero_copy_active_thresh;
/* Number of outstanding tx bytes to trigger zero copy */
extern uint32_t                 tx_bytes_out_thresh;


static unm_card_response_handler_t card_resp_handlers[UNM_P2H_OPCODE_LAST];


/*
 * Pegnet proc handling code. Located under /proc/pegnet
 */
static struct proc_dir_entry	*pegnet_dir = NULL;
static struct proc_dir_entry	*zcopy_dir = NULL;
static struct proc_dir_entry	*compressed_iov_file = NULL;
static struct proc_dir_entry	*tx_size_thresh_file = NULL;
static struct proc_dir_entry	*tx_conn_thresh_file = NULL;
static struct proc_dir_entry	*tx_bytes_out_file = NULL;
extern atomic_t g_devid;
extern index_table_t		g_listener_grp_tbl;
#define UNM_DMA_BUFFER_SIZE	(9 * 1024)

typedef enum { slosh_nocopy, slosh_docopy } unm_slosh;

/* XXX This is required to keep insmod quiet */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
long long __udivdi3(long long  op1, long long op2)
{
	return 0;
}
#endif

/*
 *
 */
int unm_create(struct net_device *netdev, struct socket *sock,
	       uint8_t rule_flags)
{
        pegnet_sock_id_t        *sid = NULL;
        int                     rv;

        if (sock->sk == NULL) {
                return (-EINVAL);
        }
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        if (!try_module_get(THIS_MODULE)) {
                return (-EINVAL);
        }
#else
        MOD_INC_USE_COUNT;
#endif

        /* Talk to card and get back sock_id_t */
        sid = host_create(netdev, PF_INET, IPPROTO_TCP, &rv);
        if (sid != NULL) {
                sid->parent = sock;
                UNM_SOCK_SET_PEGNET_SOCK_ID(sock, sid);
                sid->zero_copy_enabled = (rule_flags &
					  NETXEN_OFFLOAD_RULE_ZCOPY) ? 1 : 0;
		sid->iscsi = 0;
                rv = 0;
        } else {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
                module_put(THIS_MODULE);
#else
                MOD_DEC_USE_COUNT;
#endif
        }
        return (rv);
}

/*
 *
 */
void nx_set_route_reference(pegnet_sock_id_t *sid, uint32_t saddr,
			    uint16_t sport, uint32_t daddr, uint16_t dport)
{
	sid->lport = sport;
	sid->rport = dport;
	sid->lip_addr.v4 = saddr;
	sid->rip_addr.v4 = daddr;

	sid->route_added = 1;

	nx_print1("saddr = 0x%x, sport = %u, daddr = 0x%x, dport = %u\n",
		  saddr, sport, daddr, dport);
}

/*
 *
 */
int nx_init_connection(pegnet_sock_id_t *sid)
{
        if (sid->recv_dma.initialized) {
                return (0);
        }

        /* Check if this is a high performance flow, and act accrodingly. */
        if (atomic_read(&sid->adapter->high_perf_cnt) > 0) {
                atomic_dec(&sid->adapter->high_perf_cnt);
                sid->recv_dma.num_pcb_rxbuffers = NUM_PCB_RXBUFFERS_HP;
                sid->send_dma.num_vbuf_pages = UNM_TX_VBUF_MAX_PAGES_HP;
                sid->high_perf_flag = 1;
        } else {
                sid->recv_dma.num_pcb_rxbuffers = NUM_PCB_RXBUFFERS_LP;
                sid->send_dma.num_vbuf_pages = UNM_TX_VBUF_MAX_PAGES_LP;
                sid->high_perf_flag = 0;
        }

	return (nx_init_rxbuffers_pcb(sid));
}

/*
 *
 */
void cleanup_rx_buffers(pegnet_sock_id_t * sid)
{
        int     i;

        for (i = 0; i < pegnet_num_bufs; i++) {
                if (sid->dma[i].buffer != NULL) {
#if 0
                        pci_free_consistent(unm_device_ptr,
                                            UNM_DMA_BUFFER_SIZE,
                                            sid->dma[i].buffer,
                                            sid->dma[i].phys);
#else
                        kfree(sid->dma[i].buffer);
//// Start debug.
//            printk( "cleanup_dma_buffer: freeing buffer=%p\n", sid->dma[i].buffer ) ;
//// End debug.
#endif
                        sid->dma[i].buffer = NULL;
                }
        }
        cleanup_rxbuffers_pcb(sid);
}

void nx_schedule_task(pegnet_sock_id_t *sid)
{
        struct nx_sk_cleanup_queue      *sk_cleanup_queue;
        sk_cleanup_queue = &sid->adapter->sk_cleanup_queue;

        INIT_LIST_HEAD((&sid->tw_list));
        spin_lock_bh(&sk_cleanup_queue->lock);
        list_add_tail(&sid->tw_list, &sk_cleanup_queue->list);
        spin_unlock_bh(&sk_cleanup_queue->lock);
        NX_KC_SCHEDULE_WORK(&sid->adapter->sk_cleanup_task);
}

/*
 *
 */
int unm_release(struct socket *sock, int delay_sk_release)
{
        int                     rv = 0;
        pegnet_sock_id_t        *sid= NULL;

        sid = UNM_SOCK_GET_PEGNET_SOCK_ID(sock);
        rv = host_release(sid, delay_sk_release);
       	if (rv != 0) {
               	nx_print1("host_release failed %d\n", rv);
       	}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        module_put(THIS_MODULE);
#else
        MOD_DEC_USE_COUNT;
#endif
	if (sock->sk != NULL) {
		UNM_SOCK_SET_PEGNET_SOCK_ID(sock, NULL);
	}

        return (rv);
}

/*
 *
 */
long unm_wait_for_connect(pegnet_sock_id_t *sid, long timeo)
{
	long ret = 0;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	DEFINE_WAIT(wait);
        prepare_to_wait(&sid->active_conn_wq, &wait, TASK_INTERRUPTIBLE);
#else
        DECLARE_WAITQUEUE(wait, current);
        add_wait_queue(&sid->active_conn_wq, &wait);
        current->state = TASK_INTERRUPTIBLE;
#endif

	/*
	 * Wait for unm_proc_connect_complete() to signal us.
	 */
	spin_lock_bh(&sid->lock);
	while ((sid->active_conn_state & CONNECT_COMPLETED) == 0) {
		spin_unlock_bh(&sid->lock);
		timeo = schedule_timeout(timeo);

		if (signal_pending(current)) {
			ret = sock_intr_errno(timeo);
			goto done;
		}
		else if (!timeo)
		{
			ret = -ETIMEDOUT;
			goto done;
		}
		spin_lock_bh(&sid->lock);
		prepare_to_wait_exclusive(&sid->active_conn_wq, &wait,
				TASK_INTERRUPTIBLE);
        }
	spin_unlock_bh(&sid->lock);
done:
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        finish_wait(&sid->active_conn_wq, &wait);
#else
        current->state = TASK_RUNNING;
        remove_wait_queue(&sid->active_conn_wq, &wait);
#endif

        return (ret);
}

/*
 *
 */
int unm_connect(struct socket *sock, struct sockaddr *saddr, int len,
                int flags, int do_bind)
{
        struct sock             *sk = sock->sk;
        int                     err;
        long                    timeo = 0;
	unsigned char		state;
        pegnet_sock_id_t        *sid = NULL;

	if(sk->SK_STATE == TCP_LISTEN) {
		return (-EISCONN);
	}
        lock_sock(sk);

	sid = UNM_SOCK_GET_PEGNET_SOCK_ID(sock);

        if (saddr->sa_family == AF_UNSPEC) {
                err = unm_disconnect(sock, flags, sock_sndtimeo(sk, 0));
                sock->state = err ? SS_DISCONNECTING : SS_UNCONNECTED;
                goto out;
        }

	switch (sock->state) {
	default:
		err = -EINVAL;
		goto return_err;
	case SS_CONNECTED:
		err = -EISCONN;
		/* First time ? */
		if(sid->conn_compl == 0) {
			sid->conn_compl = 1;
			err = 0;
		}
		goto return_err;
	case SS_CONNECTING:
		err = -EALREADY;
		/* Fall out of switch with err, set for this state */
		goto return_err;
	case SS_UNCONNECTED:
		err = -EISCONN; /*Connect after listen */
		if (sk->SK_STATE != TCP_CLOSE)
			goto return_err; 
		/*
		 * Always connect as async and wait for card's initial
		 * response that moves socket state to "connecting". No
		 * other connect()s are guaranteed pending on this socket.
		 */
		err = host_connect(sid, saddr, flags | O_NONBLOCK,
				   do_bind, sk->SK_REUSE, sk->SK_RCVBUF);

		if ((err == -EALREADY) || (err == -EINPROGRESS)) {
			/*
			 * At this point, unm_proc_connect_response() has
			 * surely executed; unm_proc_connect_complete() might
			 * also have.
			 */
			spin_lock_bh(&sid->lock);
			state = sid->active_conn_state;
			BUG_ON((state & CONNECT_RESPONDED) == 0);
			if ((state & CONNECT_COMPLETED) == 0)
				sock->state = state & CONNECTING ?
				    SS_CONNECTING: SS_CONNECTED;
			spin_unlock_bh(&sid->lock);

			/* Take no action: specifically don't clean up socket */
			if (-EALREADY == err)
				goto return_err;
			break;
		} else if (0 > err)
			goto sock_error;
	}

	/* In host_connect, we have already waited for the rcv_qid, so we can
	   init rx_buffers now. We will reach this point if host_connect
	   returns 0 or -EINPROGRESS. */

	/*
	 * init_rxbuffers_pcb() sets a flag once it is initialized and
	 * does not initialize again. It is now safe to call multiple
	 * times.
	 */
	nx_init_connection(sid);

	/* That's all for async, in case of sync wait for connection
	   completetion */

	if (flags & O_NONBLOCK)
		goto out;

	/* Only sync gets this far, wait for connection completion,
	   indefinitely. */

	timeo = sock_sndtimeo(sk, 0);

        release_sock(sk);
	err = unm_wait_for_connect(sid, timeo);
	lock_sock(sk);

	if (err < 0)
		goto clobber_error; /* If the wait did not complete,
				       it should be reported as an
				       internal error */

	/* Here only if our wait had completed successfully, otherwise we would
	   have been elsewhere because of unm_wait_for_connect's error code */

	switch (sid->parent->state) {
	case SS_CONNECTED:
		sid->conn_compl = 1;
		err = 0;
		break;
	default:
		err = -1 * sk->SK_ERR;
		goto sock_error;
	}

out:

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
        /*
         * TODO: Do we really need this code or do we need it in all versions.
         * Need to check linux code to confirm.
         */
        if (err == 0 || err == -EINPROGRESS) {
                struct sockaddr_in      *sin;

                sin = (struct sockaddr_in *)saddr;
                sock->sk->daddr = sin->sin_addr.s_addr;
                sock->sk->dport = sin->sin_port;
        }
#endif

  return_err:
        release_sock(sk);
        return err;

  clobber_error:
	if (-EINTR != err && -ERESTARTSYS != err)
		err = -ECONNABORTED;
  sock_error:
        sock->state = SS_UNCONNECTED;
        if (unm_disconnect(sock, flags, timeo)) {
                sock->state = SS_DISCONNECTING;
        }
        goto out;
}

int nx_kern_disconnect(pegnet_sock_id_t	*sid)
{
        int                     rv;

	spin_lock_bh(&sid->lock);
        rv = host_conn_disconnect(sid, 1, 0, 0);
        spin_unlock_bh(&sid->lock);

        return (rv);
}
EXPORT_SYMBOL(nx_kern_disconnect);

/*
 * Input is never a listener socket; caller sets sock->state.
 */
static int unm_disconnect(struct socket *sock, int flags, long timeo)
{
	return(nx_kern_disconnect(UNM_SOCK_GET_PEGNET_SOCK_ID(sock)));
}

/*
 *
 */
int unm_accept(struct socket *sock, struct socket *newsock, int flags)
{
        pegnet_sock_id_t        *newsid;
        pegnet_sock_id_t        *sid = NULL;
	listener_group_t        *grp;
	struct list_head 	*node;
        int                     rv,found = 0;


	grp =  NX_GET_LISTENER_GROUP_FROM_SOCK(sock);
	if(grp->shutdown & RCV_SHUTDOWN){
                return (-EINVAL);
	}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        if (!try_module_get(THIS_MODULE)) {
                return (-EINVAL);
        }
#else
        MOD_INC_USE_COUNT;
#endif

        newsid = unm_host_async_accept(grp, flags, &rv);
        /* newsid = host_accept(sid, flags, &err); */

        if (NULL == newsid) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
                module_put(THIS_MODULE);
#else
                MOD_DEC_USE_COUNT;
#endif
                return (rv);
        }

        newsid->parent = newsock;
        UNM_SOCK_SET_PEGNET_SOCK_ID(newsock, newsid);

        nx_init_connection(newsid);
	newsid->conn_compl = 1;
        list_for_each(node, &grp->listener_list) {
                sid = list_entry(node, pegnet_sock_id_t, listener_list);
		if(sid->adapter == newsid->adapter) {
			found = 1;
			break;
		}
        }
	if (found == 0) {
		nx_print("No listener sid is found\n");
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
                module_put(THIS_MODULE);
#else
                MOD_DEC_USE_COUNT;
#endif
		return -1;
	}
	/* Replenish the accept just used up */
	nx_host_post_accept(sid, 1);

        return (0);
}   /* unm_accept */

/*
 *
 */
int unm_getname(struct socket *sock, struct sockaddr *uaddr, int *addr_len,
                int peer)
{
	int ret =0;
	pegnet_sock_id_t *sid;

	if(sock->sk->SK_STATE == TCP_LISTEN){
                listener_group_t *grp =
                        NX_GET_LISTENER_GROUP_FROM_SOCK(sock);
		sid = list_entry(grp->listener_list.next,pegnet_sock_id_t, listener_list);
	}
	else {
		sid = UNM_SOCK_GET_PEGNET_SOCK_ID(sock);
	}
        ret = host_getname(sid, uaddr,addr_len, peer);
	return ret;
}   /*  unm_getname */

/*
 *
 */
unsigned int unm_poll(struct socket *sock)
{
        unsigned int            rv = 0;

        if (sock->sk->SK_STATE == TCP_LISTEN) {
                listener_group_t *grp =
                        NX_GET_LISTENER_GROUP_FROM_SOCK(sock);
		spin_lock_bh(&grp->lock);
                if (!list_empty(&grp->accept_passive.accept_queue)) {
                        rv |= POLLIN | POLLRDNORM;
                }
		spin_unlock_bh(&grp->lock);
	} else {
        	pegnet_sock_id_t        *sid;
       	 	sid = UNM_SOCK_GET_PEGNET_SOCK_ID(sock);
        	spin_lock_bh(&sid->lock);
		rv = sid->poll_flags & ~POLLHUP;

		if (sid->shutdown == SHUTDOWN_MASK ||
		    (sid->recv_dma.tmp_rcv_shutdown &&
		     (sid->shutdown & SEND_SHUTDOWN))) {
			rv |= POLLHUP;
		}

		if (0 == list_empty(&(sid->recv_dma.list))) {
			rv |= (POLLIN | POLLRDNORM);
		}

		if (sid->send_dma.last_comp_valid) {
			process_last_tx_completion(sid);
		}

		if (is_tx_vbuf_allocatable(NX_ZERO_AND_REG_TX_VBUF, sid)) {
			rv |= (POLLOUT|POLLWRNORM) ;
		} else {
			sid->send_dma.need_buffer_wakeup = 1;
		}
        	spin_unlock_bh(&sid->lock);
	}


        return (rv);
}   /* unm_poll */



/*
 *
 */
int nx_kern_shutdown(pegnet_sock_id_t *sid, int listen, int how)
{
        int                     rv = 0;
	int			rc = 0;
        struct list_head 	*node, *_node;

        if (how < 0 || how > 2) {
                return (-EINVAL);
        }

	if (listen) {
        	listener_group_t *grp = sid->grp;
		grp->shutdown |= how + 1;
        	list_for_each_safe(node, _node,&grp->listener_list) {
                	sid = list_entry(node, pegnet_sock_id_t,
					 listener_list);
                	if(0 != (rv = host_shutdown(sid, how))) {
                		nx_print("Shutdown Error for Listener grp = %p"
					 " sid:%p tnic_adapter = %p "
					 "dev_port = %d err=%d\n",
					 grp,sid,sid->adapter,sid->portnum,rv);
//				nx_release_sid(sid);	 
			}
		}
 		if (list_empty(&(grp->listener_list))) {
			return rv;
 		}
		return 0;
	} else {
        	/*
         	* Adding 1 to how will make it such that bit0 set will mean RX,
		* bit1 set will mean TX.
         	*
         	* how:     RX:    0 -> 0+1 -> 0x1
         	*          TX:    1 -> 1+1 -> 0x2
         	*          Both:  2 -> 2+1 -> 0x3
         	*
         	* TBD: Need to see what is to be done for RX.
         	*
         	* Both the semaphore as well as the lock have to be taken and
		* in the correct order. This is done to lock out user sendmsg
		* call and the bottom half processing.
         	*/
	        down(&sid->send_dma.lock);
        	spin_lock_bh(&sid->lock);
	        sid->shutdown |= how + 1;	/* It has to be '|' because a
						   previous shutdown could have
						   turned on some bits. */
	        if (sid->shutdown & SEND_SHUTDOWN) {
        	        wake_up_all(&sid->send_dma.wq);
        	}

		if (sid->shutdown & SEND_SHUTDOWN) {

			if (!list_empty(&sid->send_dma.zero_wait_list)) {
				rc = nx_flush_sendpage(sid);
				UNM_HOST_PNIC_STATS_INC(DBG_CNTR_SENDPAGE_FLUSHED2);
				if (rc) {
					nx_print("Need to fix sid = %p\n", sid);
				}
			}
		}

     		if (!list_empty(&sid->send_dma.vec_wait_list) ||
		    !list_empty(&sid->send_dma.failed_list)) {
                	/*
			 * Turn off the TX shutdown condition. It will be sent
			 * when the last TX data is sent to the card.
			 */
                	how = (how + 1) & ~SEND_SHUTDOWN;
                	how -= 1;
        	}

        	spin_unlock_bh(&sid->lock);
        	up(&sid->send_dma.lock);

        	/*
		 * If the original how was 0 (tx shutdown) and we had a failure
		 * or flow control condition 'how' would have gotten
		 * decremented and will be negative in which case don't send
		 * shutdown message.
		 */
       	 	if (how >= 0) {
                	rv = host_shutdown(sid, how);
        	}
        	return (rv);
	}
}
EXPORT_SYMBOL(nx_kern_shutdown);

int unm_shutdown(struct socket *sock, int how)
{
        pegnet_sock_id_t        *sid;
	int			listen;

	if (sock->sk->SK_STATE == TCP_LISTEN) {
		listener_group_t	*grp;

		listen = 1;
		grp = NX_GET_LISTENER_GROUP_FROM_SOCK(sock);
		sid = list_entry(grp->listener_list.next, pegnet_sock_id_t,
		    listener_list);
	} else {
		listen = 0;
		sid = UNM_SOCK_GET_PEGNET_SOCK_ID(sock);
	}
	return(nx_kern_shutdown(sid, listen, how));
}

#if (UNM_X_HARDWARE==UNM_X_ASIC)
#include "af_unm_asic.c"
#else
#error "Unexpected hardware specification"
#endif

int smart_memcpy(void *dest, void *src, int size)
{
	int ret = 0;
    
	if ((unsigned long) dest < TASK_SIZE) {
		/* dest is in user space */
		//assert((unsigned long) src >= TASK_SIZE);
		ret = copy_to_user(dest, src, size);
#if defined(PEGNET_DEBUG_PRINT) && (PEGNET_DEBUG_LEVEL>1)
		debug_print_payload(pegnet_debug_mask, src, size);
#endif
	} else {
		/* dest is in kernel space */
		if ((unsigned long) src < TASK_SIZE) {
			/* src is in user space */
			ret = copy_from_user(dest, src, size);
#if defined(PEGNET_DEBUG_PRINT) && (PEGNET_DEBUG_LEVEL>1)
			debug_print_payload(pegnet_debug_mask, dest, size);
#endif
		} else {
			/* src is in kernel space */
			memcpy(dest, src, size);
		}
	}

	return ret;
}

/*
 *
 */
static void invalid_card_response(nx_tnic_adapter_t *tnic_adapter,
				  unm_msg_t *msg, pegnet_sock_id_t *sid,
				  pegnet_inprogress_t *trans)
{
        nx_print("Invalid card response\n");
}

/*
 *
 */
static void proc_create_response(nx_tnic_adapter_t *tnic_adapter,
				 unm_msg_t *msg, pegnet_sock_id_t *sid,
				 pegnet_inprogress_t *trans)
{
	host_peg_msg_hdr_t      *hdr = (host_peg_msg_hdr_t *)&(msg->body);
        create_response_t       *resp = (create_response_t *)&msg->body;
        create_response_t       *sptr;
	uint32_t		tmp_ctx;
	int			id;
        int                     replies;

	/* Retrieve the sid */
	tmp_ctx = host_get_uint32_from_peg(hdr->ctx);
	sid = host_map_tmpctx_2_sid(tnic_adapter->p_ftable,tmp_ctx);
	if (!sid) {
		printk("%s: Could not map tmp_ctx %x to sid\n",
				__FUNCTION__, tmp_ctx);
		return;
	}

	/* Retrieve the transaction to wake up */
	id = hdr->op_id;

        spin_lock(&sid->lock);
        for (trans = sid->op; trans != NULL; trans = trans->link) {
                if (trans->op_id == id) {
                        break;
                }
        }
        spin_unlock(&sid->lock);

	if (!trans) {
		printk("%s: Could not retrieve transaction for sid = %p, "
				" tmp_ctx = %x\n", __FUNCTION__, sid, tmp_ctx);
		return;
	}

	atomic_inc(&trans->replied);
        replies = atomic_read(&trans->replied);     
        if (replies != 1) {
                printk("%s: WARNING replies:%d != 1\n", __FUNCTION__, replies);
		return;
        }

        sptr = (create_response_t *)&(trans->msg.body);
        sptr->ret_code = resp->ret_code;
	if (!sptr->ret_code) {
		sptr->qid = resp->qid;
		sptr->new_ctx = resp->new_ctx;
		sptr->connId = resp->connId;

		/* update the connID table */
		host_update_connID_table(tnic_adapter->p_ftable,sid, resp->connId);
	}

        wake_up(&trans->wq);

	return;
}

/*
 *
 */
static void proc_release_response(nx_tnic_adapter_t *tnic_adapter,
				  unm_msg_t *msg, pegnet_sock_id_t *sid,
				  pegnet_inprogress_t *trans)
{
        release_response_t      *resp = (release_response_t *)&msg->body;
        spin_lock_bh(&sid->lock);
        sid->state = RELEASED_SOCK;
        sid->ret_val = host_get_uint32_from_peg(resp->ret_code);
        if(sid->ret_val == -EINPROGRESS) {
                sid->tx_buf_remaining = 1;
        }
	sid->op_needed |= CLEANUP_RX_BUF; 
	if(sid->context_type != NX_KERNEL_CONTEXT) {
		if(!sid->on_task) {
			sid->on_task = 1;
			nx_schedule_task(sid);
		}
	}
        spin_unlock_bh(&sid->lock);
        wake_up(&sid->wq);
}


/*
 *
 */
static void proc_poll_status(nx_tnic_adapter_t *tnic_adapter,unm_msg_t *msg, 
			     pegnet_sock_id_t *sid,pegnet_inprogress_t *trans)
{
        poll_status_t   *status = (poll_status_t *)&msg->body;

        spin_lock_bh(&sid->lock);
        sid->poll_flags = host_get_uint32_from_peg(status->flags);
        spin_unlock_bh(&sid->lock);

        if (sid->parent && sid->parent->sk &&
            sid->parent->sk->SK_SLEEP &&
            waitqueue_active(sid->parent->sk->SK_SLEEP)) {
                wake_up_interruptible(sid->parent->sk->SK_SLEEP);
        }
}

/*
 *
 */
static void unm_proc_conn_signal(nx_tnic_adapter_t *tnic_adapter,
				 unm_msg_t *msg, pegnet_sock_id_t *sid,
				 pegnet_inprogress_t *trans)
{
        conn_signal_t   *sig_msg = (conn_signal_t *)&msg->body;
        struct sock     *sk;

        spin_lock_bh(&sid->lock);

	if (sid && sid->parent && sid->parent->sk) {
                sk = sid->parent->sk;
	    
                if (sk && sk->SK_SOCKET && sk->SK_SOCKET->fasync_list) {
                        sock_wake_async(sk->SK_SOCKET, sig_msg->how,
                                        sig_msg->band);
                }
        }

        spin_unlock_bh(&sid->lock);
}

void nx_proc_sock_cleanup(nx_tnic_adapter_t *tnic_adapter, unm_msg_t *msg, 
			  pegnet_sock_id_t *sid, pegnet_inprogress_t *trans)
{
        spin_lock_bh(&sid->lock);
	sid->op_needed |= CLEANUP_SK;
        host_flowtbl_put_reference(sid->adapter->p_ftable, sid->connID);
	host_reset_connID_table(sid->adapter->p_ftable, sid, sid->connID);
        sid->cleanup_flag = 1;
        if (sid->release_seen ) {
                sid->cleanup_flag = 2;
		if(!sid->on_task) {
			 nx_schedule_task(sid);
		}
		if(sid->grp && sid->accept_active.synq_flag) {
			listener_group_t *grp = sid->grp;
			int err = 0;
			if(atomic_dec_and_test(&grp->child_cnt)) {
                        	kfree(grp->accept_passive.lopt);
                        	grp->accept_passive.lopt = NULL;
                        	err = host_indextbl_free_data_by_idx(&g_listener_grp_tbl,grp->index);
                        	if (err != 0) {
                                	nx_print("Free from index table failed: %d\n",err);
                        	}
			}
		}
	} else {
//                printk(KERN_ERR "%s: got invalid sid:%p\n", __func__, sid);
	}
        spin_unlock_bh(&sid->lock);
	return;
}

void nx_sk_cleanup(struct sock *sk)
{
	/* Part of host_release() */
	sk->SK_PROTINFO = NULL;

	/* Part of unm_inet_release() */
	if (sk->SK_DESTRUCT) {
		// We have waited long enough !!
		sk->SK_STATE = TCP_CLOSE;
		//ip_mc_drop_socket(sk);
		sk->SK_PROT->close(sk, 0);
	} else {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,14)
		inet_put_port(&tcp_hashinfo, sk);
#else
		tcp_put_port(sk);
#endif
		sk_free(sk);
		UNM_HOST_PNIC_STATS_DEC(DBG_CNTR_INUSE_SK);
	}
}

void nx_init_sk_cleanup_task(nx_tnic_adapter_t *tnic_adapter)
{
        INIT_LIST_HEAD(&tnic_adapter->sk_cleanup_queue.list);
        spin_lock_init(&tnic_adapter->sk_cleanup_queue.lock);

        NX_KC_INIT_WORK(&tnic_adapter->sk_cleanup_task, nx_sock_cleanup_task,
			tnic_adapter);
}


/*
 *
 */
static void nx_sock_cleanup_task(void *v)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
	struct work_struct	*_work = (struct work_struct *)v;
	nx_tnic_adapter_t	*tnic_adapter = container_of(_work,
						     struct nx_tnic_adapter_s,
						     sk_cleanup_task);
#else
	nx_tnic_adapter_t	*tnic_adapter = (nx_tnic_adapter_t *)v;
#endif
        struct list_head	free_list;
        struct list_head	*node, *n;
        pegnet_sock_id_t	*sid = NULL;
	int			flag = 0;

        INIT_LIST_HEAD(&free_list);

        spin_lock_bh(&tnic_adapter->sk_cleanup_queue.lock);
        list_for_each_safe(node, n, &tnic_adapter->sk_cleanup_queue.list) {
                sid = list_entry(node, pegnet_sock_id_t, tw_list);
                list_del(&sid->tw_list);
                list_add_tail(node, &free_list);
        }
        spin_unlock_bh(&tnic_adapter->sk_cleanup_queue.lock);

        list_for_each_safe(node, n, (&free_list)) {
                sid = list_entry(node, pegnet_sock_id_t, tw_list);
                if(!sid) {
                        printk(KERN_ERR "%s: Got empty node\n", __func__);
                        goto done;
                }

		UNM_HOST_PNIC_STATS_DEC(DBG_CNTR_SOCK_CLEANUP_CONNS);
                list_del(&sid->tw_list);
		spin_lock_bh(&sid->lock);

                if(sid->op_needed & CLEANUP_GLOBAL_RX_BUF) {
                        cleanup_rxbuffers_global(sid);
                }
        	if(sid->op_needed & CLEANUP_RX_BUF) {
                	cleanup_rx_buffers(sid);
        	}
        	if(sid->op_needed & CLEANUP_TX_BUF) {
                	cleanup_tx_vbuf_queues(sid);
        	}
                if(sid->high_perf_flag && rx_queue_empty(sid) && tx_queue_empty(sid)) {
                        atomic_inc(&sid->adapter->high_perf_cnt);
                        sid->high_perf_flag = 0;
                }	
	
		if(sid->op_needed & CLEANUP_SK)	{
			flag = 1;
		}	
                sid->on_task = 0;

                spin_unlock_bh(&sid->lock);
                if(flag == 1) {
			if(sid->sk_cleanup) {
                        	nx_sk_cleanup(sid->sk_cleanup);
                	}
                        free_pegnet_sock_id(sid);
			flag = 0;
                }
	
        }

  done:
        return;
}
                                

/*
 *
 */
static void proc_tcp_event(nx_tnic_adapter_t *tnic_adapter, unm_msg_t *msg, 
			   pegnet_sock_id_t *sid, pegnet_inprogress_t *trans)
{
        unm_tcp_event_msg_t     *event;

        event = (unm_tcp_event_msg_t *)&msg->body;

        switch (event->param.type) {
        case UNM_TCP_EVENT_FIN_ACKED:
                spin_lock_bh(&sid->lock);
                cleanup_tx_vbuf_queues(sid);
                spin_unlock_bh(&sid->lock);
                break;

        case UNM_TCP_EVENT_RST_RECEIVED:
		if (sid->accept_active.parent_sid) {
			/* We are still a child yet to be accepted */

			pegnet_sock_id_t	*parent_sid;
                        listener_group_t	*grp;

			parent_sid = sid->accept_active.parent_sid;
			grp = NX_GET_LISTENER_GROUP_FROM_SID(parent_sid);

			/* Hold the parent's lock */
			spin_lock_bh(&parent_sid->lock);

			if (grp->released) {
				spin_unlock_bh(&parent_sid->lock);
				return; /* we will be cleaned
					   up with the parent */
			}

			/* Silently discard this child */
			list_del(&sid->accept_active.list);

			spin_lock_bh(&grp->lock);
			grp->accept_passive.accepted_len--;
			spin_unlock_bh(&grp->lock);

			/* Post an accept ticket back */
			nx_host_post_accept(parent_sid, 1);
			spin_unlock_bh(&parent_sid->lock);

			host_release_unaccepted_child(sid);


			/*
			 * There is no way a sock cleanup can come before this
			 * message. If it comes in before this or is processed
			 * before this we have a serious problem.
			 *
			 * No need to free the sid here. The sid will be freed
			 * when the sock cleanup comes down.
			 */
			sid->release_seen = 1;
			sid->op_needed |= CLEANUP_GLOBAL_RX_BUF;

			UNM_HOST_PNIC_STATS_INC(DBG_CNTR_RST_ON_ACCEPT_QUEUE);

			return; /* Application does not yet know, and
				  we should not have struct socket references
				  at this point. */
		}

        case UNM_TCP_EVENT_ABORT_ON_SYN:
        case UNM_TCP_EVENT_ABORT_ON_DATA:
        case UNM_TCP_EVENT_ABORT_ON_CLOSE:
        case UNM_TCP_EVENT_ABORT_ON_MEMORY:
        case UNM_TCP_EVENT_ABORT_ON_LINGER:
        case UNM_TCP_EVENT_ABORT_DISCONNECT:
        case UNM_TCP_EVENT_KEEPALIVE_EXPIRED:
        case UNM_TCP_EVENT_RETRANSMIT_TIMEOUT:
        case UNM_TCP_EVENT_WND_PROBE_TIMEOUT:
        case UNM_TCP_EVENT_FINWAIT2_EXPIRED:
                spin_lock_bh(&sid->lock);

                sid->shutdown = SHUTDOWN_MASK;
		if (sid->parent && sid->parent->sk) {
		    sid->parent->sk->SK_ERR = (int)event->param.reason;
                }
		del_timer_sync(&sid->send_dma.sendpage_timer);
		if(sid->tx_buf_remaining == 1) {
			send_full_release_to_card(sid);
			sid->tx_buf_remaining = 0;
		}
		sid->op_needed |= CLEANUP_TX_BUF;

		if(sid->released && sid->context_type != NX_KERNEL_CONTEXT) {
			if(!sid->on_task) {
				sid->on_task = 1;
				nx_schedule_task(sid);
			}
		}
                wake_up_all(&sid->send_dma.wq);
                wake_up_all(&sid->recv_dma.wq);

                spin_unlock_bh(&sid->lock);
                break;
        }
}

/*
 *
 */
static void debug_print_tcp_event_msg(unm_msg_t *msg, pegnet_sock_id_t *sid,
				      pegnet_inprogress_t *trans)
{
        unm_tcp_event_msg_t     *event = (unm_tcp_event_msg_t *)&msg->body;
        static char		*str[] = {
		"Unknown Type",
		"FIN_ACKED",
		"RST_RECEIVED",
		"ABORT_ON_SYN",
		"ABORT_ON_DATA",
		"ABORT_ON_CLOSE",
		"ABORT_ON_MEMORY",
		"ABORT_ON_LINGER",
		"ABORT_DISCONNECT",
		"KEEPALIVE_EXPIRED",
		"RETRANSMIT_TIMEOUT",
		"WND_PROBE_TIMEOUT"
	};
	static int		count = sizeof (str) / sizeof (char *);

        printk("%s: sid is NULL for connID[%u], TCP_EVENT, %s[%u], "
	       "reason[%u/%d]\n", __FUNCTION__, (uint32_t)event->hdr.ctx,
               event->param.type < count ? str[event->param.type] : str[0],
	       event->param.type,
	       (__uint32_t)event->param.reason, (int)event->param.reason);
}

/*
 *
 */
static void proc_ifconfig_reply(nx_tnic_adapter_t *tnic_adapter,
				unm_msg_t *msg, pegnet_sock_id_t *sid,
				pegnet_inprogress_t *trans)
{
        ifconfig_response_t     *reply;

        reply = (ifconfig_response_t *)&msg->body;

	switch (reply->cmd) {
	case NX_NETDEV_ADD:
		nx_print("ifconfig %s ADD : %s\n", reply->ifname,
			 reply->ret_code ? "FAILED" : "SUCCESS");
		break;
	case NX_NETDEV_REMOVE:
		nx_print("ifconfig %s REMOVE : %s\n", reply->ifname,
			 reply->ret_code ? "FAILED" : "SUCCESS");
		break;
	case NX_NETDEV_UP:
		nx_print("ifconfig %s UP : %s\n", reply->ifname,
			 reply->ret_code ? "FAILED" : "SUCCESS");
		break;
	case NX_NETDEV_DOWN:
		nx_print("ifconfig %s DOWN : %s\n", reply->ifname,
			 reply->ret_code ? "FAILED" : "SUCCESS");
		break;
	}
}

/*
 *
 */
static void proc_resource_notice(nx_tnic_adapter_t *tnic_adapter,unm_msg_t *msg, 
			pegnet_sock_id_t *sid,pegnet_inprogress_t *trans)
{
        netxen_resource_notice_t        *notice;

        notice = (netxen_resource_notice_t *)&msg->body;
        switch (notice->type) {
        case NETXEN_TX_MEM_MS_LOW_WATER:
                printk("%s: Received NETXEN_TX_MEM_MS_LOW_WATER, data = %u\n",
                       __FUNCTION__, (uint32_t)notice->param);
                break;

        case NETXEN_TX_MEM_MS_HIGH_WATER:
                printk("%s: Received NETXEN_TX_MEM_MS_HIGH_WATER, data = %u\n",
                       __FUNCTION__, (uint32_t)notice->param);
                break;

        case NETXEN_TX_MEM_HS_LOW_WATER:
                printk("%s: Received NETXEN_TX_MEM_HS_LOW_WATER, data = %u\n",
                       __FUNCTION__, (uint32_t)notice->param);
                break;

        case NETXEN_TX_MEM_HS_HIGH_WATER:
                printk("%s: Received NETXEN_TX_MEM_HS_HIGH_WATER, data = %u\n",
                       __FUNCTION__, (uint32_t)notice->param);
                break;

        default:
                printk("%s: type = %u, data = %u\n",
                       __FUNCTION__, (uint32_t)notice->type,
		       (uint32_t)notice->param);
                break;
        }
}

/*
 *
 */
static void nx_proc_routecache_msg(nx_tnic_adapter_t *tnic_adapter,
				   unm_msg_t *msg, pegnet_sock_id_t *sid,
				   pegnet_inprogress_t *trans)
{
	nx_proc_route_invalid(tnic_adapter, (routecache_invalid_t *)msg);
}

/*
 *
 */
void register_card_response_handler(unm_card_response_handler_t *handler)
{
        card_resp_handlers[handler->opcode] = *handler;
}

EXPORT_SYMBOL(register_card_response_handler);

static unm_card_response_handler_t	initial_card_resp_handlers[] =
{
	{UNM_P2H_OPCODE_START, 0, 0,
	 invalid_card_response, NULL, "START"},
        {UNM_P2H_OPCODE_CREATE_RESPONSE, 0, 0,
	 proc_create_response, NULL, "CREATE_RESPONSE"},
        {UNM_P2H_OPCODE_RELEASE_RESPONSE, 1, 0,
	 proc_release_response, NULL, "RELEASE_RESPONSE"},
        {UNM_P2H_OPCODE_CONNECT_RESPONSE, 1, 1,
	 unm_proc_connect_response, NULL, "CONNECT_RESPONSE"},

        {UNM_P2H_OPCODE_POLL_STATUS, 1, 0,
	 proc_poll_status, NULL, "POLL_STATUS"},
        {UNM_P2H_OPCODE_RX_URGENT_POINTER, 1, 0,
	 nx_proc_rx_urg_pointer, NULL, "RX_URGENT_POINTER"},
        {UNM_P2H_OPCODE_CONN_SIGNAL, 1, 0,
	 unm_proc_conn_signal, NULL, "CONN_SIGNAL"},
        {UNM_P2H_OPCODE_SENDMSG_CONTROL, 1, 0,
	 proc_sendmsg_control, NULL, "SENDMSG_CONTROL"},

        {UNM_P2H_OPCODE_TCP_EVENT, 1, 0,
	 proc_tcp_event, debug_print_tcp_event_msg, "TCP_EVENT"},

        {UNM_P2H_OPCODE_LISTEN_RESPONSE, 1, 1,
	 unm_host_proc_listen_response, NULL, "LISTEN_RESPONSE"},
        {UNM_P2H_OPCODE_IOCTL_RESPONSE, 1, 1,
	 unm_host_proc_ioctl_response, NULL, "IOCTL_RESPONSE"},
        {UNM_P2H_OPCODE_SETSOCKOPT_RESPONSE, 1, 1,
	 unm_host_proc_setsockopt_response, NULL, "SETSOCKOPT_RESPONSE"},
        {UNM_P2H_OPCODE_GETSOCKOPT_RESPONSE, 1, 1,
	 unm_host_proc_getsockopt_response, NULL, "GETSOCKOPT_RESPONSE"},
        {UNM_P2H_OPCODE_IPTSETSOCKOPT_RESPONSE, 1, 1,
	 unm_host_proc_iptsetsockopt_response, NULL, "IPTSETSOCKOPT_RESPONSE"},
        {UNM_P2H_OPCODE_IPTGETSOCKOPT_RESPONSE, 1, 1,
	 unm_host_proc_iptgetsockopt_response, NULL, "IPTGETSOCKOPT_RESPONSE"},

        {UNM_P2H_OPCODE_CONN_SHUTDOWN_RESPONSE, 1, 1,
	 unm_host_proc_shutdown_response, NULL, "CONN_SHUTDOWN_RESPONSE"},

        {UNM_P2H_OPCODE_IFCONFIG_REPLY, 0, 0,
	 proc_ifconfig_reply, NULL, "IFCONFIG_REPLY"},
	{UNM_P2H_OPCODE_CONNECT_COMPLETE, 1, 0,
	 unm_proc_connect_complete, NULL, "CONNECT_COMPLETE"},
        {UNM_P2H_OPCODE_GET_STATS_REPLY, 0, 0,
	 nx_proc_get_stats_reply, NULL, "GET_STATS_REPLY"},
        {UNM_P2H_OPCODE_RESOURCE_NOTICE, 0, 0,
	 proc_resource_notice, NULL, "RESOURCE_NOTICE"},
        {NX_P2H_OPCODE_SOCK_CLEANUP, 1, 0,
	 nx_proc_sock_cleanup, NULL, "SOCK_CLEANUP"},
	{UNM_P2H_OPCODE_3WAY_HANDSHAKE_COMPLETE, 0, 0,
	 nx_proc_tcp_handshake_response, NULL, "3WAY_HANDSHAKE_COMPLETE"},
	{UNM_P2H_OPCODE_PRUNABLE_SK, 0, 0,
	 nx_proc_prunable_sk, NULL, "PRUNABLE_SK"},
	{UNM_P2H_OPCODE_ROUTECACHE_INVALID, 0, 0,
	 nx_proc_routecache_msg, NULL, "ROUTECACHE_INVALID"},
        {NX_P2H_OPCODE_UPDATE_CONNID, 0, 0,
         nx_proc_update_connID, NULL, "UPDATE_CONNID"}

};

/*
 *
 */
static void init_card_response_handlers(void)
{
	int	i;
	int	count;

        memset(card_resp_handlers, 0, sizeof (card_resp_handlers));

	count = (sizeof (initial_card_resp_handlers) /
		 sizeof (unm_card_response_handler_t));
	for (i = 0; i < count; i++) {
		register_card_response_handler(&initial_card_resp_handlers[i]);
	}
};

/*
 *
 */
void handleCardResponse(struct net_device *netdev,unm_msg_t *msg)
{
	host_peg_msg_hdr_t      *hdr = (host_peg_msg_hdr_t *)&(msg->body);
	int                     id;
	pegnet_sock_id_t        *sid;
	pegnet_inprogress_t     *trans;
        uint8_t                 opcode;
        uint32_t                connID;
        unm_card_response_handler_t     *handler;
        int                     replies;
	nx_tnic_adapter_t 	*tnic_adapter;

        opcode = hdr->opcode;
	tnic_adapter = nx_get_dev_tnic_adapter(netdev);

        if (opcode == 0 || opcode >= UNM_P2H_OPCODE_LAST) {
                printk("%s: Invalid opcode[%u] in message\n", __FUNCTION__,
                       opcode);
                return;
        }

        handler = &card_resp_handlers[opcode];

        if (handler->func == NULL) {
		printk("%s: Unknown opcode 0x%x\n", __FUNCTION__, opcode);
		return;
	}

        if (!handler->retreive_socket) {
                handler->func(tnic_adapter,msg, NULL, NULL);
                return;
        }

        connID = host_get_uint32_from_peg(hdr->ctx);
        sid = retreive_pegnet_sock_id(tnic_adapter->p_ftable,connID);
        if (sid == NULL) {
                if (handler->printmsg) {
                        handler->printmsg(msg, sid, NULL);
                } else {
			printk("%s: NULL sid for connID[%u]: %s[0x%x]\n",
			       __FUNCTION__, connID,
			       (handler->opcode_str != NULL ?
				handler->opcode_str : "Unknown"),
			       opcode);
                        printk("%016llx  %016llx  %016llx  %016llx\n"
                               "%016llx  %016llx  %016llx  %016llx\n",
                               ((uint64_t *)msg)[0], ((uint64_t *)msg)[1],
                               ((uint64_t *)msg)[2], ((uint64_t *)msg)[3],
                               ((uint64_t *)msg)[4], ((uint64_t *)msg)[5],
                               ((uint64_t *)msg)[6], ((uint64_t *)msg)[7]);
                }
                return;
        }

        if (!handler->retreive_transaction) {
                handler->func(tnic_adapter,msg, sid, NULL);
		goto release_reference;
        }

	id = hdr->op_id;

        spin_lock_bh(&sid->lock);
        for (trans = sid->op; trans != NULL; trans = trans->link) {
                if (trans->op_id == id) {
                        break;
                }
        }
        spin_unlock_bh(&sid->lock);

        if (trans == NULL) {
                if (handler->printmsg) {
                        handler->printmsg(msg, sid, NULL);
                } else {
			printk("%s: NULL transaction for connID[%u]: "
			       "%s[0x%x]: op_id[%u]\n",
			       __FUNCTION__, connID,
			       (handler->opcode_str != NULL ?
				handler->opcode_str : "Unknown"),
			       opcode, id);
                        printk("%016llx  %016llx  %016llx  %016llx\n"
                               "%016llx  %016llx  %016llx  %016llx\n",
                               ((uint64_t *)msg)[0], ((uint64_t *)msg)[1],
                               ((uint64_t *)msg)[2], ((uint64_t *)msg)[3],
                               ((uint64_t *)msg)[4], ((uint64_t *)msg)[5],
                               ((uint64_t *)msg)[6], ((uint64_t *)msg)[7]);
                }
		goto release_reference;
        }

        atomic_inc(&trans->replied);
        replies = atomic_read(&trans->replied);     
        if (replies == 1) {
                handler->func(tnic_adapter,msg, sid, trans);
        } else {
                printk("%s: WARNING replies:%d != 1\n", __FUNCTION__, replies);
        }

  release_reference:
	host_flowtbl_put_reference(tnic_adapter->p_ftable,connID);
        return;
}

/*
 * Allocate a receive page.
 */
int alloc_netxen_page(netxen_host_page_t *page)
{
#if	defined(NETXEN_USE_ALLOC_PAGES_HM)
#error	"Has bug in 32 bit kernel"
	page->mm_page = alloc_pages(GFP_HIGHUSER, 0);
	if (page->mm_page == NULL)
#elif	defined(NETXEN_USE_ALLOC_PAGES_LM)
#error	"Has bug in 32 bit kernel"
	page->mm_page = alloc_pages(GFP_KERNEL, 0);
	if (page->mm_page == NULL)
#else
	page->buffer = kmalloc(PAGE_SIZE, GFP_KERNEL);
	if (page->buffer == NULL)
#endif
	{
		printk("%s: Memory Allocation failed\n", __FUNCTION__);
		return (-ENOMEM);
	}

#if	defined(NETXEN_USE_ALLOC_PAGES_LM)
#error	"Has bug in 32 bit kernel"
	page->buffer = kmap(page->mm_page);
	if (page->buffer == NULL) {
		printk("%s: Memory Mapping failed\n", __FUNCTION__);
		__free_pages(page->mm_page, 0);
		return (-ENOMEM);
	}
#endif

	page->size = PAGE_SIZE;
	return (0);
}

/*
 * Free a receive page.
 */
int free_netxen_page(netxen_host_page_t *page,
		struct pci_dev *unm_device_ptr,int direction)
{
	netxen_page_unmap_pci(page, unm_device_ptr, direction);
	netxen_page_unmap_virt(page);

#if	defined(NETXEN_USE_ALLOC_PAGES_HM) || defined(NETXEN_USE_ALLOC_PAGES_LM)
#error	"Has bug in 32 bit kernel"
#if	defined(NETXEN_USE_ALLOC_PAGES_LM)
	if (page->buffer) {
		kunmap(page->mm_page);
		page->buffer = NULL;
	}
#endif	/* NETXEN_USE_ALLOC_PAGES_LM */

	if (page->mm_page) {
		__free_pages(page->mm_page, 0);
	}
#else
	if (page->buffer) {
		kfree(page->buffer);
	}
#endif
	else {
		printk("%s: can't free NULL buffer. %s.\n", __FUNCTION__,
		       (direction == PCI_DMA_FROMDEVICE) ? "RX" : "TX");
		return (1);
	}

	page->mm_page = NULL;
	page->buffer = NULL;
	page->phys = 0;

	return (0);
}

/*
 * END OF ROUTINES TO HANDLE RECEIVED DATA FROM THE CARD.
 */

extern void proc_tx_completion(struct net_device *netdev,unm_msg_t *msg);
extern void proc_tx_switch_completion(struct net_device *netdev,unm_msg_t *msg);



#if 0
void hex_dump(void *p, unsigned int l)
{
	unsigned int i = 0;
	unsigned char *c = (unsigned char *)p;

	for (; i < l; i++, c++) {
		if (!(i % 16)) {
			printk("%s%08x: ", i?"\n":"",i);
		}
		printk("%02x ", *c);
	}

	printk("\n");
}
#endif

/* This function is used to take a DMA'able address
   phys and return an address which can be used
   in smart_memcpy(). */
static void *nx_get_page_addr(struct net_device *netdev,
		unsigned long phys, __uint64_t conn_id)
{
	void                    *addr = NULL;
	nx_tnic_adapter_t       *tnic;
	pegnet_sock_id_t        *sid;
	host_tx_vbuf_t          *vbuf;
	struct list_head        *each_node;
	struct list_head        *tmp_node;
	unsigned int            i;
	unsigned long           page_start = phys & PAGE_MASK;

#if 0
	addr = phys_to_virt(phys);
#else

	tnic = nx_get_dev_tnic_adapter(netdev);
	sid = retreive_pegnet_sock_id(tnic->p_ftable, conn_id);

	if (list_empty(&sid->send_dma.sent_list)) {
		nx_print("sid %p has empty sent dma list\n", sid);
		goto release_reference;
	}

	/* Traverse the sid's sent buffers to see where phys came from.
	   This is read-only, so 'safe' was not necessary here */
	list_for_each_safe(each_node, tmp_node, &sid->send_dma.sent_list) {
		vbuf = list_entry(each_node, host_tx_vbuf_t, list);

		for (i=0; i < vbuf->num_pages; i++) {
#if 0
			printk("%s: Comparing page %d of vbuf %llx,"
					" dma addr = %llx, phys = %llx, "
					"page_start = %llx\n",
					"page_start = %llx\n",
					__FUNCTION__, i, vbuf,
					vbuf->page[i].phys, phys, page_start);
#endif

			if (page_start == vbuf->page[i].phys) {
				if (vbuf->page[i].zcopy_user_addr) {
#if 0
					addr = vbuf->page[i].zcopy_user_addr;
#endif
					/* This data can only be brought
					   over by kmapping into the kernel's
					   address space. Drop it for now. */
					addr = NULL;

					UNM_HOST_PNIC_STATS_INC(
						DBG_CNTR_LIBPCAP_ZCOPY_DROPS);
					break;
				}
				else {
					addr = netxen_page_map_virt(&vbuf->page[i]);
				}

				if (!addr) {
					break;
				}

				addr = (char *)addr + (phys - page_start);
				break;
			}

		} /* for all pages */

	} /* list for all vbufs */

release_reference:
	host_flowtbl_put_reference(tnic->p_ftable,conn_id);

#endif
	if (!addr) {
		nx_print("Unable to translate %lx to host accessible addr\n",
				phys);
	}

	return addr;
}

/*
 * Setup a received TOE skb so that it can be offered to the packet
 * subsystem.
 * Outbound skb as received from the card contains an epg-request as payload.
 * Inbound skb as received from the card contains a cloned copy of the skb
 * received on the connections flow queue.
 * WARNING: skb contents will most probably change when this function returns.
 */
int nx_setup_monitored_skb(struct net_device *netdev, 
		__uint32_t dirn, struct sk_buff *skb)
{
	int ret = 0;

	switch (dirn) {
		case NX_CAP_DIRN_IN:
			{
				unsigned short	len;
				struct ethhdr	*eth;
				eth = (struct ethhdr *)skb->mac.raw;
				len = (eth->h_proto == __constant_htons(ETH_P_8021Q)) ? VLAN_HLEN : ETH_HLEN;
				skb_push(skb,len);
				skb->pkt_type = PACKET_OTHERHOST;
				break;
			}

		case NX_CAP_DIRN_OUT:
			{
				nx_host_epg_monitor_msg_t *monitor_msg =
					(nx_host_epg_monitor_msg_t *)
					(skb->mac.raw + MBUF_ALIGNMENT_OFFSET);

				__uint32_t header_len = monitor_msg->hdr_size;

				char *setup_addr = (char *)&monitor_msg->hdr[0];

				__uint64_t header_offset = (__uint64_t)
							((unsigned long)setup_addr -
							 (unsigned long)monitor_msg);

						unsigned int i;

			/* This is an outbound skb created by TOE, so be trusting and
			   keep checks light. */
			skb_pull(skb, header_offset);
			
			/* Try to setup L2 protocol */

			if (header_len >= ETH_HLEN) {
				__uint64_t *hw_hdr = (__uint64_t *)setup_addr;
				unsigned int l2_type = (hw_hdr[1]&0xffff) >> 16;

				if (ETH_P_8021Q == l2_type) {
					skb->mac.raw = setup_addr;
					skb->protocol = ETH_P_8021Q;
				}
				else {
					skb->mac.raw = setup_addr;
					skb->protocol = ETH_P_IP;
				}

#if 0
				printk("%s: Header dump:\n", __FUNCTION__);
				hex_dump(monitor_msg, sizeof(*monitor_msg));
#endif

				setup_addr += header_len;
			}
			else {
				printk("%s: WARNING header too small (%d bytes) to"
						" determine start of L2 hdr\n",
						__FUNCTION__, header_len);
			}

			for (i=0; i<monitor_msg->num_segs; i++) {
				__uint32_t len, to_copy;
				unsigned long phys;
				void *p;
				void *copy_to;
				int rc;

				phys = monitor_msg->segs[i].addr_low |
					(((__uint64_t)
					 monitor_msg->segs[i].addr_hi) << 32);
				len = monitor_msg->segs[i].len;

#if 0
				p = phys_to_virt(phys);
#endif
				p = nx_get_page_addr(netdev, phys,
						monitor_msg->conn_id);

				/*
				printk("%s: [%d] addr %lx, len = %d, virt = %p, tail room  = %d\n", __FUNCTION__, i+1, phys, len, p,
						skb_tailroom(skb));
				 */

				to_copy = (len <= skb_tailroom(skb)) ? len : skb_tailroom(skb);
				copy_to = skb_put(skb, to_copy);

				if ((rc = smart_memcpy(copy_to, p, to_copy))) {
					nx_print("Copying %d out of %d bytes "
							"failed\n", rc,
							to_copy);
					break;
				}


#if 0
				printk("%s: Data dump:\n", __FUNCTION__);
				hex_dump(p, to_copy);
#endif

				if (to_copy != len) {
					break; /* bytes dropped, we are out of
						  room */
				}
			}
			skb->pkt_type = PACKET_OTHERHOST;
		}
		break;

	default:
		printk("%s: Unknown direction %d for skb %p\n",
				__FUNCTION__, dirn, skb);
		ret = -1;
	} /* switch */

	return ret;
}

/*
 * The tasklet to wake up sleeping processes when we get a reply
 * from our board
 */
void unm_process_pegnet_msg(struct net_device *netdev, unm_msg_t *msg)
{
        if (msg->hdr.type == UNM_MSGTYPE_CARD_RESPONSE) {
		/*
		 * We got a bunch of responses from the card. Extract the
		 * info and dispatch accordingly.
		 */
		handleCardResponse(netdev,msg);
		return;
        }

        if (msg->hdr.type == UNM_MSGTYPE_RXBUFFER) {
                /* 
                 * Received data from the card.
                 */
                nx_proc_rx_buffer(netdev,msg);
                return;
        }

        if (msg->hdr.type == UNM_MSGTYPE_SENDMSG) {
                /* 
                 * Used send buffer from the card.
                 */
                proc_tx_completion(netdev,msg);
                return;
        }

        if (msg->hdr.type == UNM_MSGTYPE_TXBUFFER_SWITCH) {
                /*
                 * Completion of switch of zcopy buffer from the card.
                 */
                proc_tx_switch_completion(netdev,msg);
                return;
        }

        /* Now check what we have to do */
	printk(KERN_ERR "%s: UNM_ISR: #5 Unexpected msg\n"
	       "0x%llx 0x%llx 0x%llx 0x%llx\n"
	       "0x%llx 0x%llx 0x%llx 0x%llx\n", __FUNCTION__,
	       msg->hdr.word, msg->body.values[0],
	       msg->body.values[1], msg->body.values[2],
	       msg->body.values[3], msg->body.values[4],
	       msg->body.values[5], msg->body.values[6]);
}

/*
 * This function is called for each flow when the table is destroyed. This
 * allows us to free the flows in the table and also notify the card to kill
 * i.e. reset each flow.
 *
 * Parameters:
 *      id      - The ID of the flow being freed.
 *      flow    - The specified flow pointer being freed.
 *
 * Return:
 *      void.
 */
void flow_tbl_destroy_flow_free(flow_id_t id, void *flow)
{
        pegnet_sock_id_t        *sid = (pegnet_sock_id_t *)flow;

        if (sid->sk_cleanup) {
		nx_sk_cleanup(sid->sk_cleanup);
		/* kfree(sid); */
        } else {
		printk("%s: Should not be here for sid = %p\n", __FUNCTION__,
		       sid);
	}
}

static int proc_read_tx_zcopy_iov_type(char *page, char **start, off_t off,
				       int count, int *eof, void *data)
{
	int len = 0;

	len = sprintf(page, "Compressed IOVs are  %s\n",
		      global_use_compressed_iov ? "ENABLED" : "DISABLED");
	return len;

}

static int proc_write_tx_zcopy_iov_type(struct file *file, const char *buffer,
					unsigned long count, void *data)
{
	int len = 0;
	char str[2];
	

	if (count == 0) {
		return 0;
	}

	if(count > 2)
		len = 2;
	else
		len = count;
	
	if(copy_from_user(str, buffer, len))
		return -EFAULT;
	

	str[len-1] = '\0';
	
	if (!strcmp(str, "1")) {
		global_use_compressed_iov = 1;

	} else if (!strcmp(str,"0")) {
		global_use_compressed_iov = 0;
	}

	return len;
}


static int proc_read_tx_zcopy_size_thresh(char *page, char **start, off_t off,
					  int count, int *eof, void *data)
{
	int len = 0;
	
	len = sprintf(page, "Zero copy Tx size threshold = %u bytes [default = 0, max = 2097152] \n",
		      tx_zero_copy_size_thresh);
	return len;

}

static int proc_write_tx_zcopy_size_thresh(struct file *file,
					   const char *buffer,
					   unsigned long count, void *data)
{
	int len = 0;
	char str[8];
	uint32_t size = 0;
	char *end;
	
	
	if (count == 0) {
		return 0;
	}

	if(count > 8)
		len = 8;
	else
		len = count;
	
	if(copy_from_user(str, buffer, len))
		return -EFAULT;
	

	str[len-1] = '\0';

	size = simple_strtoul(str, &end, 0);
	if (end == str) {
		printk("Invalid size\n");
		return len;

	}

	if (size <= 2*1024*1024) {
		tx_zero_copy_size_thresh = size;
	}


	return len;
}


static int proc_read_tx_zcopy_conn_thresh(char *page, char **start, off_t off,
					  int count, int *eof, void *data)
{
	int len = 0;
	
	len = sprintf(page, "Zero copy Tx connection threshold = %u [default = 0, max = 1000000]\n",
		      tx_zero_copy_active_thresh);
	return len;
}

static int proc_write_tx_zcopy_conn_thresh(struct file *file,
					   const char *buffer,
					   unsigned long count,
					   void *data)
{
	int len = 0;
	char str[8];
	uint32_t conn = 0;
	char *end;
	
	
	if (count == 0) {
		return 0;
	}

	if(count > 8)
		len = 8;
	else
		len = count;
	
	if(copy_from_user(str, buffer, len))
		return -EFAULT;
	

	str[len-1] = '\0';

	conn = simple_strtoul(str, &end, 0);
	if (end == str) {
		printk("Invalid size\n");
		return len;

	}

	if (conn <= 1000000) {
		tx_zero_copy_active_thresh = conn;
	}


	return len;
}


static 
int proc_read_tx_bytes_out_thresh(char *page, char **start,
				       off_t off, int count, 
				       int *eof, void *data)
{
	int len = 0;
	
	len = sprintf(page, "Zero copy Tx outstanding bytes threshold = %u "
                      "[default = 1000000, max = 4294967295 (2^32-1)]\n",
		      tx_bytes_out_thresh);
	return len;
}

static 
int proc_write_tx_bytes_out_thresh(struct file *file,
					const char *buffer,
					unsigned long count, 
					void *data)
{
	int len = 0;
	char str[8];
	uint32_t out_bytes = 0;
	char *end;
	
	
	if (count == 0) {
		return 0;
	}

	if(count > 11)
		len = 11;
	else
		len = count;
	
	if(copy_from_user(str, buffer, len))
		return -EFAULT;
	

	str[len-1] = '\0';

	out_bytes = simple_strtoul(str, &end, 0);
	if (end == str) {
		printk("Invalid size\n");
		return len;

	}

	tx_bytes_out_thresh = out_bytes;


	return len;
}

static int proc_read_max_tcp_conn(char *page, char **start,
					off_t off, int count,
					int *eof, void *data)
{
	int len = 0;

	len = sprintf(page, "%u\n", max_tcp_conn);
	return len;
}

static int proc_write_max_tcp_conn(struct file *file,
					const char *buffer,
					unsigned long count,
					void *data)
{
	int len = 0;
	char str[9];
	unsigned int max_conn;

	if (count == 0)
		return 0;

	if (count > 8)
		len = 8;
	else
		len = count;

	if (copy_from_user(str, buffer, len))
		return -EFAULT;

	str[len] = 0;

	max_conn = simple_strtoul(str, NULL, 10);
	
	if (max_conn > UNM_MAX_TCP_CONNS)
		return -EINVAL;

	printk("%s: max_conn:%d\n", __FUNCTION__, max_conn);
	max_tcp_conn = max_conn;

	return len;
}

static int proc_read_dbg_counters(char *page, char **start, off_t off,
				  int count, int *eof, void *data)
{
	int	len = 0;

#define DBG_CNTR_PRINT(BUF, LEN, STR, VARIABLE)			\
	(LEN) += sprintf((BUF) + (LEN), "    %44s\t = %d\n",	\
			 STR, UNM_HOST_PNIC_STATS_GET(VARIABLE))

        len += sprintf(page, "NetXen TOE Debug Counters\n");
        len += sprintf(page + len, "\n");

        DBG_CNTR_PRINT(page, len, "Active Connections", DBG_CNTR_ACTIVE_CONNS);
        DBG_CNTR_PRINT(page, len, "Connection Table Overflows",
		       DBG_CNTR_CONN_TBL_OVERFLOWS);
        DBG_CNTR_PRINT(page, len, "'struct sock' in use", DBG_CNTR_INUSE_SK);
        DBG_CNTR_PRINT(page, len, "TX Pages in use", DBG_CNTR_INUSE_TX_PAGES);
        DBG_CNTR_PRINT(page, len, "RX Pages in use", DBG_CNTR_INUSE_RX_PAGES);
        DBG_CNTR_PRINT(page, len, "Global RX Pages in use",
		       DBG_CNTR_INUSE_GLOBAL_RX_PAGES);
        DBG_CNTR_PRINT(page, len, "Dropped SYNs", DBG_CNTR_SYNQ_DROPPED);
        DBG_CNTR_PRINT(page, len, "SYN received for freed listener",
		       DBG_CNTR_SYN_FOR_FREED_LISTENER);
        DBG_CNTR_PRINT(page, len, "Conn complete received for freed listener",
		       DBG_CNTR_CONN_FOR_FREED_LISTENER);
        DBG_CNTR_PRINT(page, len, "Error passive connection complete",
		       DBG_CNTR_ERR_PASSIVE_CONN_COMPLETE);
        DBG_CNTR_PRINT(page, len, "Released from SYN queue",
		       DBG_CNTR_FREED_FROM_SYNQ);
        DBG_CNTR_PRINT(page, len, "Released from accept queue",
		       DBG_CNTR_FREED_FROM_ACCEPTQ);
        DBG_CNTR_PRINT(page, len, "TX Freeing In BH",
		       DBG_CNTR_TX_FREEING_IN_BH);
        DBG_CNTR_PRINT(page, len, "TX Delayed Freeing",
		       DBG_CNTR_TX_DELAY_FREEING);
        DBG_CNTR_PRINT(page, len, "Sendpage Timer Fired",
		       DBG_CNTR_SENDPAGE_TIMER_FIRED);
        DBG_CNTR_PRINT(page, len, "Sendpage Flushed 1",
		       DBG_CNTR_SENDPAGE_FLUSHED1);
        DBG_CNTR_PRINT(page, len, "Sendpage Flushed 2",
		       DBG_CNTR_SENDPAGE_FLUSHED2);
        DBG_CNTR_PRINT(page, len, "Sendpage Flushed 3",
		       DBG_CNTR_SENDPAGE_FLUSHED3);
        DBG_CNTR_PRINT(page, len, "Total Pruned", DBG_CNTR_TOTAL_PRUNED);
        DBG_CNTR_PRINT(page, len, "Total Prunables", DBG_CNTR_TOTAL_PRUNABLES);
        DBG_CNTR_PRINT(page, len, "Rst to child on accept queue",
		       DBG_CNTR_RST_ON_ACCEPT_QUEUE);
	DBG_CNTR_PRINT(page, len, "incomplete segments to pf_packet due "
			"to zero copy", DBG_CNTR_LIBPCAP_ZCOPY_DROPS);
	len += sprintf(page + len, "\n");

/*         len += sprintf(page + len, "%-60s: %d\n", "RX read timeout count", */
/* 		       val); */
        *eof = 1;
        return len;
}

static char *ip_stats_str[] = {
	"IpInReceives",
	"IpInHdrErrors",
	"IpInAddrErrors",
	"IpInNoRoutes",
	"IpInDiscards",
	"IpInDelivers",
	"IpOutRequests",
	"IpOutDiscards",
	"IpOutNoRoutes",
	"IpReasmTimeout",
	"IpReasmReqds",
	"IpReasmOKs",
	"IpReasmFails",
	"IpFragOKs",
	"IpFragFails",
	"IpFragCreates"
};

static char *tcp_stats_str[] = {
	"TcpMaxConn",
	"TcpActiveOpens",
	"TcpPassiveOpens",
	"TcpAttemptFails",
	"TcpEstabResets",
	"TcpCurrEstab",
	"TcpInSegs",
	"TcpOutSegs",
	"TcpSlowOutSegs",
	"TcpRetransSegs",
	"TcpInErrs",
	"TcpOutRsts",
	"TcpOutCollapsed",
	"TcpTimeWaitConns",
};

/*
 *
 */
static int procfs_print_ip_statistics(char *buffer, netxen_pstats_t *pstat,
				      int port)
{
	int		len = 0;
	int		ii;
	int		cnt1;
	int		cnt2;
	int		jj;
	__uint64_t	value;

	cnt1 = sizeof (ip_stats_str) / sizeof (char *);
	cnt2 = sizeof (nx_ip_mib_t) / sizeof (__uint64_t);

        len += sprintf(buffer + len, "%s", "\nIP Statistics");
        len += sprintf(buffer + len, "%s", "\n=============\n");

	if (cnt1 != cnt2) {
		len += sprintf(buffer + len, "Mismatch in number of elements "
			       "in IP string and statistics: strings[%d] != "
			       "statistics[%d]\n", cnt1, cnt2);
		return (len);
	}

	for (jj = 0; jj < cnt1; jj++) {
		len += sprintf(buffer + len, "%-20s", ip_stats_str[jj]);

		value = 0;
		for (ii = 0; ii < NETXEN_NUM_PEGS; ++ii) {
			value += ((__uint64_t *)&pstat->peg[ii].port[port].ip_statistics)[jj];
		}

		len += sprintf(buffer + len, "%10llu\n", value);
	}

	return (len);
}

/*
 *
 */
static int procfs_print_tcp_statistics(char *buffer, netxen_pstats_t *pstat,
				       int port)
{
	int		len = 0;
	int		ii,port_cnt;
	int		cnt1;
	int		cnt2;
	int		jj;
	__uint64_t	value;

	cnt1 = sizeof (tcp_stats_str) / sizeof (char *);
	cnt2 = sizeof (nx_tcp_mib_t) / sizeof (__uint64_t);

        len += sprintf(buffer + len, "%s", "\nTCP Statistics");
        len += sprintf(buffer + len, "%s", "\n==============\n");

	if (cnt1 != cnt2) {
		len += sprintf(buffer + len, "Mismatch in number of elements "
			       "in TCP string and statistics: strings[%d] != "
			       "statistics[%d]\n", cnt1, cnt2);
		return (len);
	}

	for (jj = 0; jj < cnt1; jj++) {
		len += sprintf(buffer + len, "%-20s", tcp_stats_str[jj]);
		for (port_cnt=0;port_cnt <= port;port_cnt++) {
			value = 0;
			for (ii = 0; ii < NETXEN_NUM_PEGS; ++ii) {
				value += ((__uint64_t *)&pstat->peg[ii].port[port_cnt].tcp_statistics)[jj];
			}
			len += sprintf(buffer + len, "%10llu", value);
		}
		len += sprintf(buffer + len, "\n");
	}
	len += sprintf(buffer + len, "\n");
	return (len);
}

/*
 *
 */
static int proc_read_stats(char *page, char **start, off_t off,
			   int count, int *eof, void *data)
{
	int			len = 0;
	nx_tnic_adapter_t	*tnic = (nx_tnic_adapter_t *)data;
        uint32_t		i;
	int			rv,ii;


        for (i = 0; i < NX_TNIC_MAX_PORTS; i++) {
                if (NX_GET_DEV_FROM_TNIC_ADAPTER(tnic, i)) {
			break;
		}
        }

	if (i == NX_TNIC_MAX_PORTS) {
		len += sprintf(page + len, "No device active for getting "
			       "statistics\n");
		goto done;
	}

	rv = host_get_stats(tnic, i);
	if (rv) {
		len += sprintf(page + len, "Error[%d]: Getting statistics "
			       "from card failed\n", rv);
		goto done;
	}

        for (ii = 0; ii < NX_TNIC_MAX_PORTS; ++ii) {
                if (!ii) {
			len += sprintf(page + len, "%22s"," ");
		}
              
		len += sprintf(page + len, "%7s%d  ","Port",ii);
        }

	len += procfs_print_tcp_statistics(page + len, tnic->stats.data, 3);
	len += procfs_print_ip_statistics(page + len, tnic->stats.data, 0);

  done:
        *eof = 1;
        return len;
}

/*
 *
 */
static int proc_write_stats(struct file *file, const char *buffer,
			    unsigned long count, void *data)
{
	int			len = 0;
	char			str[9];
	unsigned int		value;
	nx_tnic_adapter_t	*tnic = (nx_tnic_adapter_t *)data;
        uint32_t		i;
	int			rv;

	nx_print("adapter = %p\n", tnic);

	if (count == 0) {
		return 0;
	}

	len = (count > 8) ? 8 : count;

	if (copy_from_user(str, buffer, len)) {
		return -EFAULT;
	}

	str[len] = 0;

	value = simple_strtoul(str, NULL, 10);

	if (value != 0) {
		return -EINVAL;
	}


        for (i = 0; i < NX_TNIC_MAX_PORTS; i++) {
                if (!NX_GET_DEV_FROM_TNIC_ADAPTER(tnic,i)) {
			continue;
		}

		nx_print("Resetting stats for port <%u>.\n", i);

		rv = host_reset_stats(NX_GET_DEV_FROM_TNIC_ADAPTER(tnic,i), i);
		if (rv) {
			nx_print("Error[%d]: Reset of statistics for "
				 "port <%u> failed\n", rv, i);
		}
        }

	return (len);
}


/*
 *
 */
static int proc_read_capture(char *page, char **start, off_t off,
	               int count, int *eof, void *data)
{
       int                     len;
       nx_tnic_adapter_t       *tnic = (nx_tnic_adapter_t *)data;

       len = sprintf(page, "%u\n", tnic->packet_capture);
       return len;
}

#define NX_LIBPCAP_FW_VER_MAJ   3
#define NX_LIBPCAP_FW_VER_MIN   4
#define NX_LIBPCAP_FW_VER_SUBV  108

#define NX_MAX_PROC_CAP_DIGS  2
static int proc_write_capture(struct file *file, const char *buffer,
               unsigned long count, void *data)
{
#if 1
       int                     len;
       nx_tnic_adapter_t       *tnic = (nx_tnic_adapter_t *)data;
       char                    str[NX_MAX_PROC_CAP_DIGS + 1];
       unsigned int            value;
       int                     rv;
       struct net_device       *netdev = tnic->nx_dev;
       nic_version_t           nic_version = {0, 0, 0};

       unsigned long int       needed_fw_ver =
	       NX_LIBPCAP_FW_VER_MAJ << 24 |
	       NX_LIBPCAP_FW_VER_MIN << 16 |
	       NX_LIBPCAP_FW_VER_SUBV;

       unsigned long int       current_fw_ver;

       nx_nic_get_nic_version_number(netdev, &nic_version);

       current_fw_ver = nic_version.major << 24 |
	       nic_version.minor << 16 |
	       nic_version.sub;

       /* Is the firmware guaranteed to support libpcap? */
       if (current_fw_ver <= needed_fw_ver) {
	       nx_print("Firmware version %d.%d.%d does not support "
			       "configuring "
			       "packet capture. It should be at least "
			       "%d.%d.%d.\n",
			       nic_version.major, nic_version.minor,
			       nic_version.sub, NX_LIBPCAP_FW_VER_MAJ,
			       NX_LIBPCAP_FW_VER_MIN, NX_LIBPCAP_FW_VER_SUBV
			       + 1);
	       return -ENOTSUPP;
       }

       len = (count > NX_MAX_PROC_CAP_DIGS) ? NX_MAX_PROC_CAP_DIGS : count;

       if (copy_from_user(str, buffer, len)) {
	       nx_print("copy_from_user failed\n");
	       return -EFAULT;
       }

       str[len] = 0;

       value = simple_strtoul(str, NULL, 10);

       if (value != tnic->packet_capture) {
               rv = host_update_capture(NX_GET_DEV_FROM_TNIC_ADAPTER(tnic,0),
                               value);
               if (rv) {
                       nx_print("Error[%d]: Capture status change to %d failed"
                                       "\n", rv, value);
               }
               else {
                       tnic->packet_capture = value;
               }
       }

       return len;
#else
       nx_print("Support for libpcap is not enabled in this build\n");
       return count;
#endif
}

/*
 *
 */
static int proc_read_devices(char *page, char **start, off_t off,
			     int count, int *eof, void *data)
{
	int			len = 0;
	nx_tnic_adapter_t	*tnic = (nx_tnic_adapter_t *)data;
        uint32_t		i;

        len += sprintf(page + len, "%s", "Active Devices\t: ");

        for (i = 0; i < NX_TNIC_MAX_PORTS; ++i) {
                if (tnic->nx_port[i]) {
			len += sprintf(page + len, "%s  ",
				       tnic->nx_port[i]->name);
		}
        }

        len += sprintf(page + len, "%s", "\n");

        *eof = 1;
        return (len);
}


/*
 *
 */
static void nx_tnic_procfs_destroy_parent(void)
{
	if (tx_bytes_out_file) {
		remove_proc_entry("zcopy_bytes_out_thresh", zcopy_dir);
		tx_bytes_out_file = NULL;
	}

	if (tx_conn_thresh_file) {
		remove_proc_entry("zcopy_conn_thresh", zcopy_dir);
		tx_conn_thresh_file = NULL;
	}

	if (tx_size_thresh_file) {
		remove_proc_entry("zcopy_size_thresh", zcopy_dir);
		tx_size_thresh_file = NULL;
	}

	if (compressed_iov_file) {
		remove_proc_entry("use_compressed_IOV", zcopy_dir);
		compressed_iov_file = NULL;
	}

	if (zcopy_dir) {
		remove_proc_entry("zcopytx", pegnet_dir);
		zcopy_dir = NULL;
	}
}

/*
 * Create all the TOE related proc entries.
 */
static int nx_tnic_procfs_create_parent(void)
{
	pegnet_dir = nx_nic_get_base_procfs_dir();
	if (pegnet_dir == NULL) {
		goto error;
	}

	/*
	 * Create zero copy related directory and entries.
	 */
	zcopy_dir = proc_mkdir("zcopytx", pegnet_dir);
	if (zcopy_dir == NULL) {
		goto error;
	}

	zcopy_dir->owner = THIS_MODULE;

	/*
	 *
	 */
	compressed_iov_file = create_proc_entry("use_compressed_IOV", 0644,
						zcopy_dir);
	if (compressed_iov_file == NULL) {
		goto error;
	}
	compressed_iov_file->data = NULL;
	compressed_iov_file->read_proc = proc_read_tx_zcopy_iov_type;
	compressed_iov_file->write_proc = proc_write_tx_zcopy_iov_type;
	compressed_iov_file->owner = THIS_MODULE;

	/*
	 *
	 */
	tx_size_thresh_file = create_proc_entry("zcopy_size_thresh", 0644,
						zcopy_dir);
	if (tx_size_thresh_file == NULL) {
		goto error;
	}
	tx_size_thresh_file->data = NULL;
	tx_size_thresh_file->read_proc = proc_read_tx_zcopy_size_thresh;
	tx_size_thresh_file->write_proc = proc_write_tx_zcopy_size_thresh;
	tx_size_thresh_file->owner = THIS_MODULE;	

	/*
	 *
	 */
	tx_conn_thresh_file = create_proc_entry("zcopy_conn_thresh", 0644,
						zcopy_dir);
	if (tx_conn_thresh_file == NULL) {
		goto error;
	}
	tx_conn_thresh_file->data = NULL;
	tx_conn_thresh_file->read_proc = proc_read_tx_zcopy_conn_thresh;
	tx_conn_thresh_file->write_proc = proc_write_tx_zcopy_conn_thresh;
	tx_conn_thresh_file->owner = THIS_MODULE;	

	/*
	 *
	 */
	tx_bytes_out_file = create_proc_entry("zcopy_bytes_out_thresh", 0644,
					      zcopy_dir);
	if (tx_bytes_out_file == NULL) {
		goto error;
	}
	tx_bytes_out_file->data = NULL;
	tx_bytes_out_file->read_proc = proc_read_tx_bytes_out_thresh;
	tx_bytes_out_file->write_proc = proc_write_tx_bytes_out_thresh;
	tx_bytes_out_file->owner = THIS_MODULE;

	return (0);

  error:
	nx_tnic_procfs_destroy_parent();
	return (-ENOMEM);
}

/*
 *
 */
int nx_tnic_adapter_create_procfs(nx_tnic_adapter_t *tnic,
				  struct proc_dir_entry *adapter_dir)
{
	char tnic_dir[20];
	if (adapter_dir == NULL) {
		nx_print1("Specified adapter_dir is NULL using pegnet_dir\n");
		adapter_dir = pegnet_dir;
	}

        atomic_inc(&g_devid);
        atomic_set(&tnic->dev_id,atomic_read(&g_devid));
        sprintf(tnic_dir,"lsa_%d",atomic_read(&tnic->dev_id));
        tnic->procfs.base_dir = proc_mkdir(tnic_dir, adapter_dir);

	if (tnic->procfs.base_dir == NULL) {
		goto error;
	}

	tnic->procfs.base_dir->owner = THIS_MODULE;

	/*
	 * Create all the common variables.
	 */
	tnic->procfs.max_tcp_conn_file =
		create_proc_entry("max_offloaded_tcp_conn", 0644,
				  tnic->procfs.base_dir);
	if (tnic->procfs.max_tcp_conn_file == NULL) {
		goto error;
	}
	tnic->procfs.max_tcp_conn_file->owner = THIS_MODULE;
	tnic->procfs.max_tcp_conn_file->read_proc = proc_read_max_tcp_conn;
	tnic->procfs.max_tcp_conn_file->write_proc = proc_write_max_tcp_conn;
	tnic->procfs.max_tcp_conn_file->data = tnic;

	/*
	 *
	 */
	tnic->procfs.dbg_counters_file =
		create_proc_entry("dbg_counters", 0644, tnic->procfs.base_dir);
	if (tnic->procfs.dbg_counters_file == NULL) {
		goto error;
	}
	tnic->procfs.dbg_counters_file->owner = THIS_MODULE;
	tnic->procfs.dbg_counters_file->read_proc = proc_read_dbg_counters;
	tnic->procfs.dbg_counters_file->write_proc = NULL;
	tnic->procfs.dbg_counters_file->data = tnic;

	/*
	 *
	 */
	tnic->procfs.stats_file = create_proc_entry("stats", 0644,
						    tnic->procfs.base_dir);
	if (tnic->procfs.stats_file == NULL) {
		goto error;
	}
	tnic->procfs.stats_file->owner = THIS_MODULE;
	tnic->procfs.stats_file->read_proc = proc_read_stats;
	tnic->procfs.stats_file->write_proc = proc_write_stats;
	tnic->procfs.stats_file->data = tnic;

	/*
	 * Lists the actual ethernet devices.
	 */
	tnic->procfs.devices_file = create_proc_entry("devices", 0644,
						      tnic->procfs.base_dir);
	if (tnic->procfs.devices_file == NULL) {
		goto error;
	}
	tnic->procfs.devices_file->owner = THIS_MODULE;
	tnic->procfs.devices_file->read_proc = proc_read_devices;
	tnic->procfs.devices_file->write_proc = NULL;
	//tnic->procfs.devices_file->write_proc = proc_write_devices;
	tnic->procfs.devices_file->data = tnic;

	tnic->procfs.capture = create_proc_entry("capture", 0644,
			tnic->procfs.base_dir);
	if (tnic->procfs.capture == NULL) {
		goto error;
	}
	tnic->procfs.capture->owner = THIS_MODULE;
	tnic->procfs.capture->read_proc = proc_read_capture;
	tnic->procfs.capture->write_proc = proc_write_capture;
	tnic->procfs.capture->data = tnic;

	return (0);

  error:
	nx_tnic_adapter_destroy_procfs(tnic, adapter_dir);
	return (-ENOMEM);
}

/*
 *
 */
void nx_tnic_adapter_destroy_procfs(nx_tnic_adapter_t *tnic,
				    struct proc_dir_entry *adapter_dir)
{
	char tnic_dir[20];
	if (adapter_dir == NULL) {
		nx_print1("Specified adapter_dir is NULL using pegnet_dir\n");
		adapter_dir = pegnet_dir;
	}

	if (tnic->procfs.max_tcp_conn_file) {
		remove_proc_entry("max_offloaded_tcp_conn",
				  tnic->procfs.base_dir);
		tnic->procfs.max_tcp_conn_file = NULL;
	}

	if (tnic->procfs.dbg_counters_file) {
		remove_proc_entry("dbg_counters", tnic->procfs.base_dir);
		tnic->procfs.dbg_counters_file = NULL;
	}

	if (tnic->procfs.stats_file) {
		remove_proc_entry("stats", tnic->procfs.base_dir);
		tnic->procfs.stats_file = NULL;
	}

	if (tnic->procfs.devices_file) {
		remove_proc_entry("devices", tnic->procfs.base_dir);
		tnic->procfs.devices_file = NULL;
	}

	if (tnic->procfs.capture) {
		remove_proc_entry("capture", tnic->procfs.base_dir);
		tnic->procfs.capture = NULL;
	}
	
	if (tnic->procfs.base_dir) {
                sprintf(tnic_dir,"lsa_%d",atomic_read(&tnic->dev_id));
                remove_proc_entry(tnic_dir, adapter_dir);
		tnic->procfs.base_dir = NULL;
	}
}

#define NX_NIC_SYMBOLS_NOT_DEFINED (nx_nic_is_netxen_device == NULL || \
		                nx_nic_register_lsa == NULL || \
		                nx_nic_get_adapter_revision_id == NULL || \
		                nx_nic_unregister_lsa == NULL || \
		                nx_nic_get_tnic_adapter == NULL || \
		                nx_nic_get_device_port == NULL || \
		                nx_nic_get_device_ring_ctx == NULL || \
		                nx_nic_send_tnic_msg == NULL || \
		                nx_nic_get_pcidev == NULL || \
		                nx_nic_cmp_adapter_id == NULL || \
		                nx_nic_get_base_procfs_dir == NULL || \
		                nx_nic_get_tnic_version_number == NULL || \
		                nx_nic_get_nic_version_number == NULL)

/*
 *
 */
static int __init unm_init(void)
{
        int     rv;

	if(NX_NIC_SYMBOLS_NOT_DEFINED) {
		nx_print("Nx nic module is not loaded\n");
		return(-ENOENT);
	}


        UNM_HOST_PNIC_STATS_INIT();
        init_card_response_handlers();

	if ((rv = nx_tnic_procfs_create_parent()) != 0) {
		printk("unm_init: Procfs initialization failed %d\n", rv);
		return rv;
	}

	pstack_init();

	nx_lsa_iscsi_init();
#if 0
	{
		pegnet_sock_id_t	u;

		nx_print1("sizeof pegnet_sock_id_t            = %u\n",
			  sizeof (pegnet_sock_id_t));
		nx_print1("sizeof hostRxBuffer_t              = %u\n",
			  sizeof (hostRxBuffer_t));
		nx_print1("sizeof netxen_host_page_t          = %u\n",
			  sizeof (netxen_host_page_t));
		nx_print1("sizeof recv_dma                    = %u\n",
			  sizeof (u.recv_dma));
		nx_print1("sizeof skb_shared_info             = %u\n",
			  sizeof (struct skb_shared_info));
		nx_print1("sizeof sk_buff                     = %u\n",
			  sizeof (struct sk_buff));
		nx_print1("sizeof SMP_CACHE_BYTES             = %u\n",
			  SMP_CACHE_BYTES);
	}
#endif

	/* nx_init_rx_global_buffers(); */

#ifdef CONFIG_UNM_NETFILTER
	unm_iptable_init();
#endif

        printk("ENABLING LINUX TOE\n") ;

	return 0;
}

void nx_cleanup_all_tnics(void);
static void __exit unm_done(void)
{
	/*
	 * First make sure all our scheduled tasks have completed.
	 */
	nx_print("In unm done");
	nx_lsa_iscsi_fini();
	nx_cleanup_all_tnics();
	pstack_exit();

	flush_scheduled_work();

	nx_tnic_procfs_destroy_parent();
}

module_init(unm_init);
module_exit(unm_done);

#if defined(KGDB) && !defined(DEBUG)
void NOINLINE inline prefetch(const void *x) { }        /* Might be compiled without inlines */
#endif
