/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
/*
 * Copyright and CVS version tags.
 */

/*
 * This has the global functions for the corresponding 'c' file.
 */

#ifndef	_NX_LISTEN_H
#define	_NX_LISTEN_H

#include <linux/types.h>
#include <linux/sched.h>
#include <asm/current.h>
#include <linux/interrupt.h>
#include <linux/spinlock.h>
#include <linux/net.h>
#include <linux/in.h>

#define INDEX_TABLE_SIZE				0x2000


#define	UNM_INVALID_INDEX_ID                     ((uint32_t)0xffffffff)
#define UNM_INAVLID_INDEX_IDX			((uint32_t)0xffffffff)


/*
 * The index ID struct - now a simple uint32_t
 */
typedef union {
	uint32_t	idx;
} index_id_t;

/*
 * The free ID list structure.
 */
typedef struct {
	index_id_t	*ids;
	uint32_t	max;
	uint32_t	head;
	uint32_t	tail;
} index_free_ids_t;

/*
 * The index element structure.
 */
typedef struct {
	index_id_t	id;
	void		*data;
	atomic_t	refcnt;
} index_tbl_entry_t;

/*
 * The index table structure.
 */
typedef struct {
	index_tbl_entry_t	*p_data;
	uint32_t		max;
	index_free_ids_t        free_list; /* this table tracks indices into
					      the indexes array */
	spinlock_t		lock;
	int 			nr_indexes;
} index_table_t;


/*
 * This function is used to allocate space for the index table. It allocates the
 * maximum IDs and index pointers. Does not allocate the index itself.
 *
 * Parameters:
 *	None.
 *
 * Return:
 *	0		- If allocated successfully.
 *	ENOMEM		- If out of memory.
 */
int host_indextbl_init(index_table_t *tcp_tbl);

/*
 * This function is used to free space allocated for the flow table.
 *
 * Parameters:
 *	None.
 *
 * Return:
 *	None.
 */
void host_indextbl_destroy(index_table_t *tcp_tbl);

/*
 * This function is used to allocate a index from the index table.
 * Returns data contained in the index if allocated else returns NULL.
 *
 * Parameters:
 *	index_id: Index of data in host's index-id table	
 *
 * Return:
 *	The data that was allocated.
 *	NULL	- If allocation fails.
 */
void *host_indextbl_allocate_index(index_table_t *tcp_tbl, uint32_t *idx);

/*
 * This function is used to free a flow from the table
 *
 *
 * Parameters:
 *	id		- The ID of the flow to free.
 *
 * Return:
 *	0		- If released successfully.
 *	-1		- If the flow is not found.
 */

int host_indextbl_free_data_by_idx(index_table_t *tcp_tbl, uint32_t idx);

void host_indextbl_get_reference(index_table_t *tcp_tbl, uint32_t id);
void host_indextbl_put_reference(index_table_t *tcp_tbl, uint32_t id);


/*
 * This function is used to get data from the index table.
 *
 * Parameters:
 *	idx		- The index of the data to get.
 *
 * Return:
 *	The data that was found.
 *	NULL		- If there is no data for the index.
 */
void *host_indextbl_get_data_by_idx(index_table_t *tcp_tbl, uint32_t idx);

int host_indextbl_get_nr_indexes(index_table_t *tcp_tbl);

/* Here starts listen functions */

#define NX_SYNQ_HSIZE	512
#define NX_MAX_SYNQ_LEN	1024
typedef struct {
	int			qlen;
	uint32_t		hash_rnd;
	struct list_head	youngsters;
	struct list_head	syn_table[NX_SYNQ_HSIZE];
} nx_lopt_t;

typedef struct {
	struct list_head        sid_list;
	spinlock_t              lock;
} nx_synq_prun_requests_t;

struct listener_group_s {
        struct list_head listener_list;
        struct socket	*parent;        /* pointer to parent socket */
        /*
         * Async accept requests are queued in posted_queue. The maximum
         * depth of this queue is the socket backlog value.
         *
         * When new connections are accepted they are removed from the
         * posted_queue and added to the accepted_queue.
         *
         * When the application makes an accept call then the connection in
         * accepted_queue is removed and handed over to the application. Then
         * a new accept request is posted to the posted_queue.
         */
#define UNM_POSTED_ACCEPT_MAX           PEGNET_MAX_SAVED_ACCEPTS
#define UNM_ACCEPT_STACK_MASK           0x01
        /*
         * This structure is only used by the passive listening
         * socket and not by the active connection socket.
         */
        struct {
                nx_lopt_t               *lopt;
                struct list_head        accept_queue;
                struct list_head        pruning;
                uint32_t                pruning_active;
                /*
                 * For now just alternate between host and offload
                 * stack when accepting connections. nxt_stack is
                 * either 0 or 1.
                 */
                uint8_t                 nxt_stack;
                uint32_t                accepted_len;
        } accept_passive;
        spinlock_t lock;
	uint8_t released;
	uint8_t shutdown;
        uint32_t index;
	atomic_t child_cnt;
	unsigned short	sport;
	struct in_addr	saddr;
	unsigned short	backlog;
	char		reuseflg;
        unsigned int	rcvbuf;
};

int nx_listen(struct listener_group_s *grp, int backlog, int do_bind);
int __nx_release_listener_sock(struct listener_group_s *grp);
int nx_release_listener_sock(struct socket *sock);

#endif	/* _INDEX_TABLE_H */

