/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
/*
 * Copyright (C) 2003 - 2007 NetXen, Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *                            
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *                                   
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA  02111-1307, USA.
 *
 * The full GNU General Public License is included in this distribution
 * in the file called LICENSE.
 *
 * Contact Information:
 *    info@netxen.com
 * NetXen,
 * 3965 Freedom Circle, Fourth floor,
 * Santa Clara, CA 95054
 */
#include <linux/sockios.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>

#include <linux/version.h>
#include <linux/autoconf.h>
#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
#define MODVERSIONS
#endif
#if defined(MODVERSIONS) && !defined(__GENKSYMS__)
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#include <linux/modversions.h>
#endif
#endif

#include <linux/errno.h>
#include <linux/types.h>
#include <linux/socket.h>
#include <linux/in.h>
#include <linux/unistd.h>
#include <linux/pagemap.h>
#include <linux/pci.h>
#include <linux/net.h>
#include <linux/poll.h>
#include <linux/skbuff.h>
#include <net/protocol.h>
#include <net/udp.h>
#include <net/ip.h>
#include <net/tcp.h>

#include <linux/inetdevice.h>
#include <linux/if.h>
#include <linux/if_vlan.h>

#include <unm_inc.h>
#include <pegnet_inc.h>
#include "nic_cmn.h"
#include "hostInterface.h"
#include "flow_table.h"
#include "unm_host_pnic_types.h"


extern int iscsi_toe_data_ready(struct sock *sk, struct sk_buff *skb);
extern int iscsi_register_toe(int set_tcp_offload(struct sock *sk, int iscsi));
extern int iscsi_unregister_toe(void);

#pragma	weak	iscsi_toe_data_ready
#pragma	weak	iscsi_register_toe
#pragma	weak	iscsi_unregister_toe

static void nx_lsa_iscsi_rcv(pegnet_sock_id_t *sid, hostRxBuffer_t *rbuf);
static int nx_lsa_iscsi_set_tcp_offload(struct sock *sk, int iscsi);

void nx_lsa_iscsi_init(void)
{
	if (nx_is_iscsi_lsa_capable()) {
		printk(KERN_ALERT "LSA found a compatible iSCSI initiator "
		       "module\n");
		nx_kern_register_iscsi_handler(nx_lsa_iscsi_rcv);
		iscsi_register_toe(nx_lsa_iscsi_set_tcp_offload);
	} else {
		printk(KERN_ALERT "LSA did not find a compatible iSCSI "
		       "initiator module\n");
	}
}

void nx_lsa_iscsi_fini(void)
{
	if (nx_is_iscsi_lsa_capable()) {
		iscsi_unregister_toe();
	}
}

int nx_is_iscsi_lsa_capable(void)
{
	int	rv = 1;
	static int	message_printed = 0;

	if (iscsi_toe_data_ready == NULL || iscsi_register_toe == NULL ||
	    iscsi_unregister_toe == NULL) {
		if (!message_printed) {
			printk("No iSCSI module with TCP offload support.\n");
			message_printed = 1;
		}
		rv = 0;
	}
	return (rv);
}

static int nx_lsa_iscsi_set_tcp_offload(struct sock *sk, int iscsi)
{

	if (sk->SK_FAMILY == AF_INET && sk->SK_PROTINFO) {
		pegnet_sock_id_t	*sid;

		sid = (pegnet_sock_id_t *)(sk->SK_PROTINFO);
		sid->iscsi = iscsi ? 1 : 0;
		return (0);
	}

	return (1);
}

static void nx_lsa_iscsi_rcv(pegnet_sock_id_t *sid, hostRxBuffer_t *rbuf)
{
	struct sk_buff	*skb;
	int		ii;
	int		old_data_len;
	int		old_len;
	int		old_truesize;
        int		total_bytes;
	struct page	*page;
	int		sz;


	skb = dev_alloc_skb(0);
	if (skb == NULL) {
		printk("%s: Could not alloc skb\n", __FUNCTION__);
		return;
	}
	/*
	 * Now, add the pages.
	 */

	old_data_len = skb->data_len;
	old_len = skb->len;
	old_truesize = skb->truesize;

	total_bytes = rbuf->pageBytesRemaining + rbuf->bufferBytesRemaining;

	for (ii = 0; total_bytes && ii < rbuf->numPages; ii++) {

		sz = (total_bytes < rbuf->page[ii].size ?
		      total_bytes : rbuf->page[ii].size);
		total_bytes -= sz;

		page = rbuf->page[ii].mm_page;
#if	!defined(NETXEN_USE_ALLOC_PAGES_HM) && !defined(NETXEN_USE_ALLOC_PAGES_LM)
		page = virt_to_page(rbuf->page[ii].buffer);
#endif
		skb_fill_page_desc(skb, ii, page, 0, sz);
	}
	if (total_bytes) {
		nx_print("BUG bytes still remain[%d]\n", total_bytes);
	}


	total_bytes = rbuf->pageBytesRemaining + rbuf->bufferBytesRemaining;
	skb->data_len += total_bytes;
	skb->len += total_bytes;
	skb->truesize += total_bytes;

	if (skb->len) {
		int	rc;
		int	len;

		len = skb->len;
		rc = iscsi_toe_data_ready(sid->parent->sk, skb);
		if (rc != len) {
			nx_print("ISCSI did not consume all the data: %d/%d\n",
				 rc, len);
		}
	}

	/*
	 * Remove the fragments that we added.
	 */
	skb_shinfo(skb)->nr_frags = 0;
	skb->len = old_len;
	skb->data_len = old_data_len;
	skb->truesize = old_truesize;

	/*
	 * All done, free up the skb.
	 */
	kfree_skb(skb);
}
