/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
#include <unm_inc.h>
#include <pegnet_inc.h>

#include <net/sock.h>
#include <linux/ctype.h>
#include <linux/poll.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/pagemap.h>
#include <linux/random.h>
#include <linux/version.h>
#include <linux/if_vlan.h>
#include <linux/inetdevice.h>
#include <linux/if.h>
#include "hostInterface.h"
#include "flow_table.h"
#include "unm_host_pnic_types.h"
#include "kernel_compatibility.h"
#include <linux/tcp.h>
#include "unm_pstats.h"

extern void cleanup_rx_buffers(pegnet_sock_id_t * sid);

int sysctl_tcp_stdurg=0;
extern int   global_use_compressed_iov ;

#if defined(PEGNET_DEBUG_PRINT)

volatile long pegnet_host_rpc_debug_level = PEGNET_DEBUG_LEVEL;

#define KERN_LEVEL KERN_ERR

# define DPRINTF(level, args) {                         \
  if (level < PEGNET_DEBUG_LEVEL_MAX) {                 \
    if (level <= pegnet_host_rpc_debug_level) {     	\
        if (PEGNET_DEBUG_RPC & pegnet_debug_mask) { 	\
            printk args;                            	\
        }                                           	\
    }                                               	\
  }                                               	\
}

void debug_print_payload(int mask, const char* data, int bytes)
{
    if (PEGNET_DEBUG_PAYLOAD & mask)
    {
        int bcount = 0;
        long start = (long)data & ~(sizeof(__uint64_t) - 1);
        __uint64_t* word = (__uint64_t*)start;  /* aligned start */
        const char* bdata = (const char*)start;  /* aligned start */

        printk(KERN_LEVEL "[host] payload %p for %d bytes\n", data, bytes);
        if (0 == bytes) return;  /* no point beating a dead horse */
        bytes = (bytes + sizeof(__uint64_t) - 1) & ~(sizeof(__uint64_t) - 1);
        while (bcount < bytes)
        {
            char buffer[120] = {0};
            int index, blen = sprintf(buffer, "%8p: ", word);

            blen += sprintf(buffer + blen, "%016llx ", *word++);
            if ((bcount + sizeof(__uint64_t)) < bytes)
                blen += sprintf(buffer + blen, "%016llx ", *word++);

            while (blen < 48) blen += sprintf(buffer + blen, " ");

            blen += sprintf(buffer + blen, " *");
            for (index = 0; index < (2 * sizeof(__uint64_t)); ++index)
            {
                char c = bdata[bcount++];
                if (!isprint(c)) c = '.';
                blen += sprintf(buffer + blen, "%c", c);
                if (bcount == bytes) break;
            }
            printk(KERN_LEVEL "    %s*\n", buffer);
        }
    }
}  /* debug_print_payload */

#else  /* !defined(PEGNET_DEBUG_PRINT) */

#define DPRINTF(...)

#endif /* defined(PEGNET_DEBUG_PRINT) */


/*
 * Used to reset the firmware software resources. Currently to drain the
 * global buffer queue.
 */
int nx_tnic_reset_fw(struct net_device *dev)
{
        tnic_reset_fw_t	*req;
        unm_msg_t	msg;

        req = (tnic_reset_fw_t *)&msg.body;
        memset(req, 0, sizeof (*req));

        UNM_MSGQ_HDR(msg.hdr, PEGNET_H2C_MAJOR, PEGNET_H2C_MINOR,
                     TNIC_RESET_FW);

        return (unm_send_h2p_msg(dev, &msg, sizeof (*req)));
}

/*
 * Used to configure/un-configure the pegnet device interface. Sets the IP
 * address, MAC address, etc.
 */
int host_ifconfig(struct net_device *dev, struct in_ifaddr *ifa,
		  int port, uint16_t cmd, int logical)
{
        ifconfig_request_t      *req;
        unm_msg_t               msg;
	uint64_t		hwaddr = 0;
	ip_addr_t		ipaddr = { v6: {0, 0, 0, 0}};
	nx_tnic_adapter_t       *tnic_adapter;

	if (ifa) {
		ipaddr.v4 = ifa->ifa_address;
	}

        req = (ifconfig_request_t *)&msg.body;
        memset(req, 0, sizeof (*req));
        req->hdr.opcode = UNM_H2P_OPCODE_IFCONFIG_REQUEST;
        tnic_adapter = nx_get_dev_tnic_adapter(dev);

       if (!tnic_adapter)
               return -1;
 	
        req->ring_ctx = tnic_adapter->ring_ctx;
	if (ifa) {
		strncpy(req->ifname, ifa->ifa_label, UNM_IFCONFIG_IFNAMSIZ);
	} else {
		strncpy(req->ifname, dev->name, UNM_IFCONFIG_IFNAMSIZ);
	}
        req->port = port;
        req->cmd = cmd;
	req->logical = logical ? 1 : 0;
        req->family = UNM_IFCONFIG_ADDR_V4;	/* Only v4 supported */

	if (dev->priv_flags & IFF_802_1Q_VLAN) {
		req->vlanid = VLAN_DEV_INFO(dev)->vlan_id;
		req->vlan_set = 1;
	}

	if (dev->dev_addr) {
		memcpy(&hwaddr, dev->dev_addr, dev->addr_len);
	}

        req->hwaddr = hwaddr;
        req->ipaddr = ipaddr;

        UNM_MSGQ_HDR(msg.hdr, PEGNET_H2C_MAJOR, PEGNET_H2C_MINOR,
                     HOST_REQUEST);

        return (unm_send_h2p_msg(dev, &msg, sizeof (*req)));
}  /* host_ifconfig */


static int wait_reply_timeout(pegnet_sock_id_t *sock, pegnet_inprogress_t *op,
                               int task_state)
{
        long    timeo;
        int rv = 0;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	DEFINE_WAIT(wait);
#else
        DECLARE_WAITQUEUE(wait, current);
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        prepare_to_wait(&op->wq, &wait, task_state);
#else
        add_wait_queue(&op->wq, &wait);
        /* current->state = TASK_UNINTERRUPTIBLE; */
	set_current_state(task_state);
#endif
        timeo = HOST_TIMEOUT;

        while (atomic_read(&op->replied) == 0) {
                if (!timeo) {
                        rv = -ETIMEDOUT;
                        goto done;
		} else if(task_state == TASK_INTERRUPTIBLE &&
                          signal_pending(current)) {
                        rv = -EINTR;
                        goto done;
                }
                spin_unlock_bh(&sock->lock);
		timeo = schedule_timeout(timeo);
                spin_lock_bh(&sock->lock);
        }

done :
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        finish_wait(&op->wq, &wait);
#else
        current->state = TASK_RUNNING;
        remove_wait_queue(&op->wq, &wait);
#endif
        return rv;
}

static int host_wait_reply(pegnet_sock_id_t *sid, pegnet_inprogress_t *op)
{
        int     rv = 0;

        spin_lock_bh(&sid->lock);

        if (op->type == op_release) {
                rv = wait_reply_timeout(sid, op, TASK_UNINTERRUPTIBLE);
        } else {
                rv = wait_reply_timeout(sid, op, TASK_INTERRUPTIBLE);

                if (rv == -EINTR && signal_pending(current)) {
                        DPRINTF(2, (KERN_LEVEL "%s: interrupted\n", me));

                        rv = wait_reply_timeout(sid, op, TASK_UNINTERRUPTIBLE);
                        if (rv) {
                                printk("%s: Signal received.  Due to bug, "
                                       "temporarily disabling cancel message\n"
                                       "     NEED TO FIX.\n", __FUNCTION__);
                                rv = -EINTR;
                        }
		}
	}
        spin_unlock_bh(&sid->lock);

        return (rv);
}  /* host_wait_reply */


static int host_begin_operation(pegnet_sock_id_t *sid, pegnet_inprogress_t *op,
                                host_op_type_t op_type)
{
        pegnet_inprogress_t     *list;
        __uint8_t               start;
        int                     err = 0;

        DPRINTF(4, (KERN_LEVEL "%s(sid:%p, op:%p, op_type:%d): "
                    "sid->{op:%p, op_id:%d, ctx:%x}\n",
                    __FUNCTION__, sid, op, op_type, sid->op, (int)sid->op_id,
                    sid->ctx.context));

        atomic_set(&op->replied, 0);
        /* op->sid = sid; */
        init_waitqueue_head(&op->wq);
        spin_lock_bh(&sid->lock);
        start = sid->op_id;
        op->type = op_type;

        if (op_type != op_release) {
                if (sid->released) {
                        /*
                         * Don't allow any new msgs after RELEASE is started.
                         */
                        err = -ENOTSOCK;
                        spin_unlock_bh(&sid->lock);
                        goto done;
                }
        }

        /* insure that the current operation id isn't in use */
        do {
                sid->op_id += 1;
                if (sid->op_id == start) {
                        printk("%s:%d: more than 255 operations!",
                               __FUNCTION__, __LINE__);
                        return (-ENOMEM);
                }
                for (list = sid->op; NULL != list; list = list->link) {
                        DPRINTF(8, (KERN_LEVEL "%s: list:%p->op_id:%d\n",
				    __FUNCTION__, list, list->op_id));

                        if (list->op_id == sid->op_id) {
                                break;
                        }
                }
        } while (list != NULL);
        op->op_id = sid->op_id;
        op->link = sid->op;  	/* link current list to this operation */
        sid->op = op;   /* and put this operation at the head of the list */
        DPRINTF(8, (KERN_LEVEL "%s: sid->{op:%p, link:%p}\n",
                    __FUNCTION__, sid->op, op->link));

        spin_unlock_bh(&sid->lock);

  done:
        if (err) {
                DPRINTF(2, (KERN_LEVEL "%s: return(err:%d); sid->op_id:%d;}\n",
                            __FUNCTION__, err, (int) sid->op_id));
        } else {
                DPRINTF(5, (KERN_LEVEL "%s: return(err:%d); sid->op_id:%d;}\n",
                            __FUNCTION__, err, (int) sid->op_id));
        }
        return(err);
}  /* host_begin_operation */


static int host_end_operation(
    pegnet_sock_id_t *sid, pegnet_inprogress_t *op)
{
    UNUSED static char *me = "host_end_operation";
    pegnet_inprogress_t *this;
    pegnet_inprogress_t *prev = NULL;
    pegnet_inprogress_t *last_op;
    int rv;

    DPRINTF(6, (KERN_LEVEL "%s(sid:%p, op:%p): { sid->op:%p\n",
			   me, sid,    op,       sid->op));

    rv = host_wait_reply(sid, op);  /* wait for the reply */

    spin_lock_bh(&sid->lock);

    /*
     * Remove op from list
     */
    for (this = sid->op; this != NULL; this = this->link)
    {
        if (this == op)
        {
            if (prev == NULL) {
		sid->op = this->link;
	    } else {
		prev->link = this->link;
	    }
            break;
        }
        prev = this;
    }

    if (this == NULL) {
            printk("%s: Could not find operation\n", __FUNCTION__);
    }

    last_op = sid->op;

    if ((last_op != NULL) && (last_op->link == NULL)) {
#if 0
	/*
	 * There is just 1 operation left on the queue.
	 * If it's a release operation then wake it up;
	 * it's been waiting for the rest of the op's to finish.
	 *
	 * Currently disabled till this code path can be exercised.
	 */
	if(last_op->type == op_release) {
	    printk(KERN_ERR "%s: Calling wake_up(&last_op:%p->wq);\n", me, last_op);
	    wake_up(&last_op->wq);
	}
#else
        printk(KERN_ERR "%s: Do we need to activate host_release wait code?\n", me ) ;
        printk(KERN_ERR "    Need to understand. Note how to produce.\n");
        printk(KERN_ERR "    last_op->type=%d\n", last_op->type);
#endif
    }

    spin_unlock_bh(&sid->lock);

    DPRINTF(7, (KERN_LEVEL "%s: return(rv:%d); }\n",
			    me,        rv));
    return rv;
}  /* host_end_operation */

/*
 *
 */

inline void init_redirector_fields(pegnet_sock_id_t *sid)
{
        int i;

        sid->lock = SPIN_LOCK_UNLOCKED;
        for(i=0; i<pegnet_num_bufs; i++) {
                init_MUTEX(&(sid->dma[i].lock));
                sid->dma[i].buffer = NULL;
                sid->dma[i].phys = 0;
        }

	init_MUTEX(&(sid->recv_dma.lock));
	init_waitqueue_head( &(sid->recv_dma.wq) );
	INIT_LIST_HEAD( &(sid->recv_dma.list) ) ;
	sid->recv_dma.rcv_shutdown = 0 ;
	sid->recv_dma.readOffset = 0 ;

	init_MUTEX(&(sid->send_dma.lock));
	init_waitqueue_head( &(sid->send_dma.wq) );
	init_waitqueue_head(&sid->active_conn_wq);
	INIT_LIST_HEAD(&sid->send_dma.alloc_list);
	INIT_LIST_HEAD(&sid->send_dma.sent_list);
	INIT_LIST_HEAD(&sid->send_dma.vec_wait_list);
	INIT_LIST_HEAD(&sid->send_dma.failed_list);
	INIT_LIST_HEAD(&sid->send_dma.zcopy_switch_list);
	INIT_LIST_HEAD(&sid->send_dma.zero_wait_list);

	init_timer(&sid->send_dma.sendpage_timer);
	
	sid->send_dma.vec_cnt = UNM_SEND_BUF_MAX_VECS;
	sid->send_dma.refid = 1;
        sid->send_dma.num_vbuf_pages = UNM_HOST_TX_PAGES;
	sid->send_dma.tstamp_flag = 0;

#ifdef	PEGNET_NIC
        INIT_LIST_HEAD(&(sid->listener_list));
#endif	/* PEGNET_NIC */
}

/*
 * Allocate the required memory and also allocate a new connection ID. Then
 * initialize all the required fields.
 */
pegnet_sock_id_t *alloc_pegnet_sock_id(struct net_device *netdev)
{
        pegnet_sock_id_t        *sid;
	uint32_t		flow_idx;
	int			port;
	nx_tnic_adapter_t *tnic_adapter = nx_get_dev_tnic_adapter(netdev);
        port = nx_get_device_port(netdev);
        if (port < 0) {
                nx_print("%s: Not NetXen card\n",netdev->name);
		return NULL;
        }	

	if (tnic_adapter == NULL){
		nx_print("Adapter is null. This should not happen\n");
		return NULL;
	}
//	printk("%s:g_adapter=%p tcp_tbl=%p\n",__FUNCTION__,&g_tnic_adapter,g_tnic_adapter.p_ftable);
	sid = (pegnet_sock_id_t *)host_flowtbl_allocate_flow(tnic_adapter->p_ftable,&flow_idx);
	if (sid) {
		memset(sid, 0, sizeof(pegnet_sock_id_t));
		sid->connID = UNM_INVALID_CONNID;
		sid->flow_id = flow_idx;
		sid->tmp_ctx = flow_idx;
		sid->adapter = tnic_adapter;
		sid->portnum = port;
		init_redirector_fields(sid);
		UNM_HOST_PNIC_STATS_INC(DBG_CNTR_ACTIVE_CONNS);
	} else {
		nx_print1("No entry in table\n");
		sid = NULL;
		UNM_HOST_PNIC_STATS_INC(DBG_CNTR_CONN_TBL_OVERFLOWS);
	}

        return (sid);
}

/*
 * Frees a specified pegnet_sock_id_t. First it is removed from the host table
 * and then it is kfreed.
 */
void free_pegnet_sock_id(pegnet_sock_id_t *sid)
{
        int     rv;

	if (sid->route_added) {
		nx_host_rt_key_t	key;

		key.saddr = sid->lip_addr.v4;
		key.daddr = sid->rip_addr.v4;
		nx_route_remove(sid->adapter, &key);
		sid->route_added = 0;
	}


	/* Free a connID table entry if we are holding one */
	if (UNM_INVALID_CONNID != sid->connID) {
		rv = host_flowtbl_free_flow(sid->adapter->p_ftable,sid->connID);
	} else {
		rv = host_flowtbl_free_flow_by_flowidx(sid->adapter->p_ftable,sid->flow_id);
	}

        if (rv != 0) {
                printk("%s: free from flow table failed: %d\n",
		       __FUNCTION__, rv);
        }

        UNM_HOST_PNIC_STATS_DEC(DBG_CNTR_ACTIVE_CONNS);
}

#if 0
/*
 *
 */
static int wait_for_tx_completions(pegnet_sock_id_t *sid)
{
	if (is_tx_waiting_for_completion(sid)) {
		DECLARE_WAITQUEUE(wait, current);

		nx_print1("Wait for TX buf: SL[%u], FL[%u], VL[%u], ZSL[%u]\n",
			  list_empty(&sid->send_dma.sent_list),
			  list_empty(&sid->send_dma.failed_list),
			  list_empty(&sid->send_dma.vec_wait_list),
			  list_empty(&sid->send_dma.zcopy_switch_list));

		add_wait_queue(&sid->send_dma.wq, &wait);
		while (is_tx_waiting_for_completion(sid)) {
			set_current_state(TASK_UNINTERRUPTIBLE);
			spin_unlock_bh(&sid->lock);
			schedule();
			spin_lock_bh(&(sid->lock));
		}
		set_current_state(TASK_RUNNING);
		remove_wait_queue(&sid->send_dma.wq, &wait);
	}

        return (0);
}
#endif

/*
 *
 */
static long nx_wait_for_release(pegnet_sock_id_t *sid, uint8_t linger,
                                 uint32_t lingertime)
{
        long ret = 0;
        long timeo = 0;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        DEFINE_WAIT(wait);
#else
        DECLARE_WAITQUEUE(wait, current);
#endif

        spin_lock_bh(&sid->lock);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        prepare_to_wait(&sid->wq, &wait, TASK_INTERRUPTIBLE);
#else
        add_wait_queue(&sid->wq, &wait);
        current->state = TASK_INTERRUPTIBLE;
#endif
        timeo = (linger) ? lingertime:MAX_SCHEDULE_TIMEOUT;
        /*
         * If the socket is non blocking then the call still waits for the
         * reply from the card.
         */
        while (RELEASING_SOCK == sid->state) {
                prepare_to_wait_exclusive(&sid->wq, &wait,
                                TASK_INTERRUPTIBLE);
                if(RELEASED_SOCK == sid->state) {
                        break;
                }
                if (signal_pending(current)) {
                        ret = sock_intr_errno(timeo);
                        nx_print1("Signal is pending for sid=%p connID=%d \n", sid,sid->connID);
                        break;
                }
                spin_unlock_bh(&sid->lock);
                timeo = schedule_timeout(timeo);
                spin_lock_bh(&sid->lock);
                if (!timeo) {
                        ret = -ETIMEDOUT;
                        break;
                }
        }

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        finish_wait(&sid->wq, &wait);
#else
        current->state = TASK_RUNNING;
        remove_wait_queue(&sid->wq, &wait);
#endif
        spin_unlock_bh(&sid->lock);
        return (ret);
}

/*
 *
 */
int send_full_release_to_card(pegnet_sock_id_t *sid)
{
        int                     rv = 0;
        pegnet_queue_addr_t     remote = sid->qid;
        release_request_t       *host_req;
        unm_msg_t               msg;

        host_req = (release_request_t *)&msg.body;
        memset(host_req, 0, sizeof (*host_req));

        host_req->force = 1;
        host_req->hdr.ctx = sid->ctx.context;
        UNM_MSGQ_HDR(msg.hdr, remote.major, remote.minor, RELEASE_REQUEST);
        rv = unm_send_h2p_msg(NX_GET_DEV_FROM_TNIC_ADAPTER(sid->adapter,sid->portnum),
                        &msg, sizeof(release_request_t));
        return (rv);
}

/*
 *
 */
int host_release(pegnet_sock_id_t *sid, int delay_sk_release)
{
	int			rv = 0;
	int			rc = 0;
	pegnet_queue_addr_t	remote = sid->qid;
	release_request_t	*host_req;
	unm_msg_t               msg;
        struct sock             *sk = sid->parent->sk;
	struct socket           *sock = sid->parent;	
	int			flag = 0;


	if (null_queue(remote)) {
		rv  = -ENOTSOCK;
		goto done;
	}

	spin_lock_bh(&sid->lock);

 	if (!list_empty(&sid->send_dma.zero_wait_list)) {
		UNM_HOST_PNIC_STATS_INC(DBG_CNTR_SENDPAGE_FLUSHED3);

		rc = nx_flush_sendpage(sid);
		if (rc) {
			nx_print("Need to fix sid = %p\n", sid);
		}
	}

	/*
	 * Stop new processes from sleeping on resources,
	 * clobbers all other flag.
	 */
	sid->released = 1;
	sid->state = RELEASING_SOCK;
        sid->shutdown = SHUTDOWN_MASK;
	sid->context_type = NX_KERNEL_CONTEXT;
	spin_unlock_bh(&sid->lock);

	down( &(sid->dma[pegnet_send_buf].lock));
	down( &(sid->dma[pegnet_sopt_buf].lock));

	wake_up_interruptible(sid->parent->sk->SK_SLEEP);

	init_waitqueue_head(&sid->wq);
	host_req = (release_request_t *)&msg.body;
        memset(host_req, 0, sizeof (*host_req));

        /* Applications forget to leave groups before exiting */

        /* ip_mc_drop_socket(sk); */

        /*
         * If linger is set, we don't return until the close
         * is complete.  Otherwise we return immediately. The
         * actually closing is done the same either way.
         *
         * If the close is due to the process exiting, we never
         * linger..
         */
        host_req->linger = SK_LINGER_FLAG(sk) ? 1 : 0;
        host_req->lingertime = sk->SK_LINGERTIME;
        host_req->process_exiting = (current->flags & PF_EXITING) ? 1 : 0;
        host_req->last_read_seq_num = 0;
        host_req->force = 0;

        if (!list_empty(&sid->send_dma.failed_list) ||
            !list_empty(&sid->send_dma.vec_wait_list)) {
                /* Check zero linger _after_ checking for unread data. */
                host_req->tx_bufs_pending = 1;
        }

	if (!list_empty(&(sid->recv_dma.list))) {
		/* force an RST if unread data is pending */
		host_req->reset = 1;
	}

	host_req->hdr.ctx = sid->ctx.context;
	UNM_MSGQ_HDR(msg.hdr, remote.major, remote.minor, RELEASE_REQUEST);
	rv = unm_send_h2p_msg_sleep(NX_GET_DEV_FROM_TNIC_ADAPTER(sid->adapter,
								 sid->portnum),
				    &msg, 1);
        if (rv == 0) {
                rv = nx_wait_for_release(sid, host_req->linger,host_req->lingertime);   /* wait for response
                                                           post-conditioning */
                if (rv == 0) {
                        rv = sid->ret_val;
                }
                wake_up_interruptible(sid->parent->sk->SK_SLEEP); /* One last
                                                                     try */
        } else {
                rv = -EAGAIN;
                nx_print("Unable to post release down to card sid=%p connID=%d err= %d\n", sid,sid->connID,rv);
                goto done;
        }

        if (rv == -EINPROGRESS) {
                nx_print1("ret from previous release = %d sid=%p connID=%d \n", rv,sid,sid->connID);
        }

#if 0
	if(sk->SK_STATE != TCP_FIN_WAIT2) {
		/* don't clean connID etc. in case of TIME_WAIT */
		free_pegnet_sock_id(sid);
	}
#endif
        spin_lock_bh(&sid->lock);
	/*
	 * cleanup_flag == 1. This means the sock cleanup message has come in
	 *		      from the card. In this case we can free
	 *		      the whole socket normally since the card
	 *		      has released it.
	 * cleanup_flag == 2.  This means the sid is queued in the to
	 *		      be killed list. This state can not happen
	 *		      here since release_seen and tw are not set yet.
	 */
	if (sid->cleanup_flag == 0) {
	 	/* Listener sk and connecting sk don't put the sid in task */
	 	if (delay_sk_release) {
        	 	sid->sk_cleanup = sk;
			sock->sk = NULL;
		}
                sid->release_seen = 1;
                UNM_HOST_PNIC_STATS_INC(DBG_CNTR_SOCK_CLEANUP_CONNS);
	} else if (sid->cleanup_flag == 1) {
		flag = 1;
	} else {
         	nx_print("cleanup_flag=%u, sid:%p. INVALID STATE\n",
			 sid->cleanup_flag, sid);
        }

	if(sid->op_needed & CLEANUP_RX_BUF) {
		cleanup_rx_buffers(sid);
	}
	if(sid->op_needed & CLEANUP_TX_BUF) {
		cleanup_tx_vbuf_queues(sid);
	}
	if(sid->high_perf_flag && rx_queue_empty(sid) && tx_queue_empty(sid)) {
		atomic_inc(&sid->adapter->high_perf_cnt);
		sid->high_perf_flag = 0;
	}	
	sid->parent = NULL;
	sid->context_type = NX_NON_KERNEL_CONTEXT;
	// Take care of possible races
	if(rv  == -EINPROGRESS && !is_tx_waiting_for_completion(sid) 
			&& (sid->tx_buf_remaining == 1)) {
		send_full_release_to_card(sid);
		sid->tx_buf_remaining = 0;
	}

	spin_unlock_bh(&sid->lock);

	if (flag == 1) {
		free_pegnet_sock_id(sid);
	}

  done:
	return(rv);
} /* host_release */

/*
** This is the caller side of the sock_create rpc. In practice,
** this code or some like it will end up in the host kernel, hence
** the function names.
*/
pegnet_sock_id_t *host_create(struct net_device *netdev, int family,
			      int protocol, int *pegnet_rval)
{
        create_request_t        *host_req;
        unm_msg_t               msg;
        pegnet_inprogress_t     op;
        create_response_t       *resp;
        int                     rv;
        pegnet_sock_id_t        *sock;
	nx_tnic_adapter_t       *tnic_adapter;

        sock = alloc_pegnet_sock_id(netdev);
        if (sock == NULL) {
                rv = -ENOMEM;
                goto done;
        }

        rv = host_begin_operation(sock, &op, op_create);  /* pre-condition the
                                                             op-block */
        if (rv != 0) {
                free_pegnet_sock_id(sock);
                sock = NULL;  /* seen as a failure */
                goto done;
        }

        host_req = (create_request_t *)&msg.body;
        host_req->hdr.ctx = host_encode_uint32_for_peg(sock->tmp_ctx);
        host_req->hdr.op_id = op.op_id;
        tnic_adapter = nx_get_dev_tnic_adapter(netdev);
        host_req->ring_ctx = tnic_adapter->ring_ctx;
        host_req->family = host_encode_uint16_for_peg((__uint16_t)family);
        host_req->protocol = host_encode_uint16_for_peg((__uint16_t)protocol);

        UNM_MSGQ_HDR(msg.hdr, PEGNET_H2C_MAJOR, PEGNET_H2C_MINOR,
                     CREATE_REQUEST);
        rv = unm_send_h2p_msg(netdev, &msg, sizeof (create_request_t));
        if (rv == 0) {
                rv = host_end_operation(sock, &op);  /* post-conditioning */
                if (rv == 0) {
                        resp = (create_response_t *)&op.msg.body;
                        rv = resp->ret_code;
                        if (rv == 0) {
                                host_assign_qid_from_peg(sock->qid, resp->qid);
				sock->connID =
					host_get_uint32_from_peg(resp->connId);
                                sock->ctx.context = resp->new_ctx; /* and it's
                                                                      ctx */
                                goto done;
                        }
                }
        } else {
                rv = -ENOMEM;
        }
        free_pegnet_sock_id(sock);
        sock = NULL;  /* seen as a failure */

done:
        /* caller wants return value from pegnet_socket() */
        if (pegnet_rval) {
                *pegnet_rval = rv;
        }
        DPRINTF(4, (KERN_LEVEL "%s: return(sock:%p); }\n", me, sock));
        return(sock);

}  /* host_create */

/*
 *
 */
int host_connect(pegnet_sock_id_t *sock, struct sockaddr *uaddr,
                 int flags, int do_bind, char reuse, int rcvbuf)
{
        int                     rv = 0;
        pegnet_inprogress_t     op;
        pegnet_queue_addr_t     remote = sock->qid;
        connect_request_t       *host_req;
        unm_msg_t               msg;
        struct sockaddr_in      *inaddr;

        if (null_queue(remote)) {
                rv = -ENOTSOCK;
                goto done;
        }

        rv = host_begin_operation(sock, &op, op_connect);   /* pre-condition
                                                               the op-block */
        if (rv != 0) {
                goto done;
        }

        host_req = (connect_request_t *)&msg.body;

        switch (uaddr->sa_family) {
	case AF_INET:
                inaddr = (struct sockaddr_in *)uaddr;
                host_req->family = inaddr->sin_family;
                host_req->rip_addr.v4 = sock->rip_addr.v4;
                host_req->lip_addr.v4 = sock->lip_addr.v4;
                host_req->rport = sock->rport;
                host_req->lport = sock->lport;
                host_req->flags = flags;
		sock->reported_rcvbuf = rcvbuf;
		host_req->rcvbuf = nx_h2p_uint32((sock->reported_rcvbuf / 2));
		host_req->do_bind = do_bind ? 1 : 0;
                break;

        default:
                rv = -EINVAL;
                goto done;
        }

	host_req->reuseflg = reuse;

        host_req->hdr.ctx = sock->ctx.context;
        host_req->hdr.op_id = op.op_id;

        UNM_MSGQ_HDR(msg.hdr, remote.major, remote.minor, CONNECT_REQUEST);
        rv = unm_send_h2p_msg(NX_GET_DEV_FROM_TNIC_ADAPTER(sock->adapter,sock->portnum),
			&msg, sizeof (connect_request_t));
        if (rv == 0) {
                rv = host_end_operation(sock, &op);  /* post-conditioning */
                if (rv == 0) {
                        connect_response_t	*resp;

                        resp = (connect_response_t *)&(op.msg.body);
                        rv = resp->ret_code;
                        host_assign_qid_from_peg(sock->rcvBuf_qid,
                                                 resp->rcvBuf_qid);
                }
        } else {
                rv = -ENOMEM;
        }

  done:
        return rv;
}  /* host_connect */

/*
 *
 */
void unm_proc_connect_response(nx_tnic_adapter_t *tnic_adapter,unm_msg_t *msg, 
			pegnet_sock_id_t *sid,pegnet_inprogress_t *trans)
{
        connect_response_t      *resp = (connect_response_t *)&msg->body;
	connect_response_t      *copy_to = (connect_response_t *)
						&(trans->msg.body);

        spin_lock_bh(&sid->lock);

	/* Extract the fields that we need, no need to check no. of
	   replies, we are alone. */
	copy_to->ret_code = resp->ret_code;
	copy_to->rcvBuf_qid = resp->rcvBuf_qid;

	trans->msg.hdr.type = msg->hdr.type;
	trans->msg.hdr.src_major = msg->hdr.src_major;
	trans->msg.hdr.src_minor = msg->hdr.src_minor;

	if (-EINPROGRESS == copy_to->ret_code ||
            -EALREADY == copy_to->ret_code) {
		sid->active_conn_state = CONNECT_RESPONDED | CONNECTING;
        } else if (!copy_to->ret_code) {
		/* will not happen at present */
		sid->active_conn_state = CONNECT_RESPONDED | CONNECTED;
        }
	/* else do not transition the state here, when unm_connect
	   decides to clean-up, we will move to disconnecting etc. */

	/* Wake up host_connect() */
	wake_up(&trans->wq);

        spin_unlock_bh(&sid->lock);
	return;
}

/*
 *
 */
void unm_proc_connect_complete(nx_tnic_adapter_t *tnic_adapter,unm_msg_t *msg, 
			pegnet_sock_id_t *sid,pegnet_inprogress_t *trans)
{
	connect_complete_t      *cc = (connect_complete_t *)&msg->body;

	if (sid->released)
		return;

	spin_lock_bh(&sid->lock);

	sid->active_conn_state = CONNECT_RESPONDED | CONNECT_COMPLETED |
	    ((cc->card_error == 0) ? CONNECTED : 0);
	wake_up_interruptible(&sid->active_conn_wq);

	/*
	 * RDMA temporary.
	 */
	if (sid->parent == NULL)
		goto out;

	sid->parent->state = (cc->card_error == 0) ? SS_CONNECTED :
	    SS_UNCONNECTED;
	sid->parent->sk->SK_ERR_SOFT = cc->card_error;
	sid->parent->sk->SK_ERR = cc->card_error;
	wake_up_interruptible(sid->parent->sk->SK_SLEEP);

	/*
	 * Notify parent of the connection state. Is this needed?
	 */
	if (sid->parent->sk && sid->parent->sk->SK_DATA_READY)
		sid->parent->sk->SK_STATE_CHANGE(sid->parent->sk);

out:
	spin_unlock_bh(&sid->lock);
	return;
}


/*
 * This function processes the accept_response from the card. If successful
 * it will put the entry into accepted queue and wake up the waiting user
 * context. It also posts new accept request to the card.
 */
void unm_host_proc_accept_error_resp(nx_tnic_adapter_t *tnic_adapter,unm_msg_t *msg, 
			pegnet_sock_id_t *sock,pegnet_inprogress_t *trans)
{
        accept_error_response_t       *resp;

        resp = (accept_error_response_t *)&msg->body;

	printk("%s: error response received: %d\n", __FUNCTION__,
               resp->ret_code);

	return;
}  /* unm_host_proc_accept_error_resp */


/*
 *
 */
int host_getname(pegnet_sock_id_t *sid, struct sockaddr *uaddr, int *addrlen,
                 int peer)
{
        struct sockaddr_in *sin = (struct sockaddr_in *)uaddr;

        sin->sin_family = AF_INET;
        if (peer) {
                if (!sid->rport) {
                        return (-ENOTCONN);
                }
                if (sid->parent->state != SS_CONNECTED &&
                    sid->parent->state != SS_CONNECTING) {
                        return (-ENOTCONN);
                }
                sin->sin_port = sid->rport;
                sin->sin_addr.s_addr = sid->rip_addr.v4;
        } else {
/*                 if (sid->parent->state != SS_CONNECTED && */
/*                     sid->parent->state != SS_CONNECTING) { */
/*                         __uint32_t      addr = sid->parent->sk->sk_rcv_saddr; */
/*                         if (!addr) */
/*                                 addr = sid->parent->sk->sk_saddr; */
/*                         sin->sin_port = sid->parent->sk->sk_sport; */
/*                         sin->sin_addr.s_addr = addr; */
/*                 } else { */
                        sin->sin_port = sid->lport;
                        sin->sin_addr.s_addr = sid->lip_addr.v4;
/*                 } */
        }
        *addrlen = sizeof(*sin);
        return (0);
}  /* host_getname */

/*
 * In case sk specific ioctl, ioctl_sk should be non-zero
 */
int host_ioctl(pegnet_sock_id_t* sock, int cmd, unsigned long val, int ioctl_sk)
{
        int                     rv;
        pegnet_inprogress_t     op;
        unm_msg_t               msg;
        pegnet_queue_addr_t     remote = sock->qid;
        ioctl_request_t         *req;
        ioctl_response_t        *resp;
        int                     major_q, minor_q;

        DPRINTF(4, (KERN_LEVEL "%s(sock:%p, cmd:0x%x, val:0x%lx) {\n", 
                    __FUNCTION__, sock, cmd, val));

        if (ioctl_sk != 0 && null_queue(remote)) {
                rv = -ENOTSOCK;
                goto done;
        }

        rv = host_begin_operation(sock, &op, op_ioctl); /* pre-condition the
                                                           op-block */
        if (rv != 0) {
                goto done;
        }
        req = (ioctl_request_t *)&msg.body;
        req->hdr.opcode = UNM_H2P_OPCODE_IOCTL_REQUEST;
        req->hdr.op_id = op.op_id;
        req->hdr.ctx = sock->ctx.context;
        req->ctx = host_encode_uint32_for_peg(sock->connID);
        req->cmd = host_encode_uint32_for_peg(cmd);
        req->arg = host_encode_uint64_for_peg(val);
        if(ioctl_sk != 0) {
                major_q = remote.major;
                minor_q = remote.minor;
        } else {
                major_q = PEGNET_H2C_MAJOR;
                minor_q = PEGNET_H2C_MINOR;
        }

        UNM_MSGQ_HDR(msg.hdr, major_q, minor_q, HOST_REQUEST);

        rv = unm_send_h2p_msg(NX_GET_DEV_FROM_TNIC_ADAPTER(sock->adapter,sock->portnum),
			&msg, sizeof (ioctl_request_t));
        if (rv == 0) {
                rv = host_end_operation(sock, &op);  /* post-conditioning */
                if (rv == 0) {
                        resp = (ioctl_response_t *)&(op.msg.body);
                        rv = resp->ret_code;
                }
        } else {
                rv = -ENOMEM;
        }

  done:
        DPRINTF(4, ("%s: return(rv:%d); }\n", me, rv));
        return (rv);
}  /* host_ioctl */

/*
 * Called when a response is received for an ioctl request. Stores the result
 * and wakes up the process that made the request.
 *
 * Parameters:
 *      msg     - The message received from the card.
 *      sid     - The sock context retreived for this response.
 *      trans   - The transaction of the corresponding request.
 *
 * Returns:
 *      Void.
 */
void unm_host_proc_ioctl_response(nx_tnic_adapter_t *tnic_adapter,unm_msg_t *msg, 
			pegnet_sock_id_t *sid,pegnet_inprogress_t *trans)
{
        ioctl_response_t        *resp = (ioctl_response_t *)&msg->body;
        ioctl_response_t        *req;

        DPRINTF(5, ("%s: Called\n", __FUNCTION__));

        req = (ioctl_response_t *)&(trans->msg.body);
        req->ret_code = host_get_uint32_from_peg(resp->ret_code);
        wake_up(&trans->wq);
}

/*
 *
 */
int host_shutdown(pegnet_sock_id_t *sid, int how)
{
        int                             rv = -ENOTSOCK;
        pegnet_inprogress_t             op;
        unm_msg_t                       msg;
        pegnet_queue_addr_t             remote = sid->qid;
        conn_disconnect_request_t       *req;
        conn_disconnect_response_t      *resp;

        DPRINTF(4, ("%s(sid:%p, how:%d) {\n", me, sid, how));

        if(sid == NULL) {
                goto done;		/* return -ENOTSOCK */
        }

        if (null_queue(remote)) {
                goto done;
        }

        rv = host_begin_operation(sid, &op, op_shutdown);  /* pre-condition
                                                              the op-block */
        if (rv != 0) {
                goto done;
        }
        req = (conn_disconnect_request_t *)&msg.body;
        req->hdr.opcode = UNM_H2P_OPCODE_CONN_SHUTDOWN_REQUEST;
        req->hdr.op_id = op.op_id;
        req->hdr.ctx = sid->ctx.context;
        req->abortive = 0;
        req->tx_shutdown = ((how + 1) & SEND_SHUTDOWN) ? 1 : 0;
        req->rx_shutdown = ((how + 1) & RCV_SHUTDOWN) ? 1 : 0;

        UNM_MSGQ_HDR(msg.hdr, remote.major, remote.minor, HOST_REQUEST);

        rv = unm_send_h2p_msg(NX_GET_DEV_FROM_TNIC_ADAPTER(sid->adapter,sid->portnum),
			&msg, sizeof (*req));
        if (rv == 0) {
                rv = host_end_operation(sid, &op);  /* post-conditioning */
                if (rv == 0) {
                        resp = (conn_disconnect_response_t *)&(op.msg.body);
                        rv = resp->ret_code;
                }
        } else {
                rv = -EAGAIN;
        }

  done:
        DPRINTF(4, ("%s: return(rv:%d); }\n", __FUNCTION__, rv));
        return (rv);
}  /* host_shutdown */

/*
 * Called when a response is received for an shutdown request. Stores the
 * result and wakes up the process that made the request.
 *
 * Parameters:
 *      msg     - The message received from the card.
 *      sid     - The sock context retreived for this response.
 *      trans   - The transaction of the corresponding request.
 *
 * Returns:
 *      Void.
 */
void unm_host_proc_shutdown_response(nx_tnic_adapter_t *tnic_adapter,unm_msg_t *msg, 
			pegnet_sock_id_t *sid,pegnet_inprogress_t *trans)
{
        conn_disconnect_response_t      *resp;
        conn_disconnect_response_t      *req;

        DPRINTF(5, ("%s: Called\n", __FUNCTION__));

        resp = (conn_disconnect_response_t *)&msg->body;
        req = (conn_disconnect_response_t *)&(trans->msg.body);
        req->ret_code = host_get_uint32_from_peg(resp->ret_code);
        wake_up(&trans->wq);
}

/*
 * This call should not take the lock nor should it do any wait. It can be
 * called from soft irq. It will set appropriate state if necessary.
 */
int host_conn_disconnect(pegnet_sock_id_t *sid, int abortive, int tx, int rx)
{
	int                             rv = 0;
	pegnet_queue_addr_t             remote = sid->qid;
	conn_disconnect_request_t       *req;
        unm_msg_t                       msg;

	/* printk("%s: sid:%p, ctx:%u\n", __FUNCTION__, sid, sid->ctx.context); */

	if (null_queue(remote)) {
		return (-ENOTSOCK);
	}

        memset(&msg, 0, sizeof (msg));
	UNM_MSGQ_HDR(msg.hdr, remote.major, remote.minor, DISCONNECT_OFFLOAD);
	req = (conn_disconnect_request_t *)&msg.body;
	req->hdr.ctx = sid->ctx.context;
	req->hdr.op_id = 0;
        req->abortive = abortive;
        req->tx_shutdown = tx;
        req->rx_shutdown = rx;

	rv = unm_send_h2p_msg(NX_GET_DEV_FROM_TNIC_ADAPTER(sid->adapter,sid->portnum),
			&msg, sizeof(conn_disconnect_request_t));
	if (rv != 0) {
                rv = -ENOBUFS;
        }

	return (rv);
}

/*
 *
 */
int host_setsockopt(pegnet_sock_id_t *sock, int level, int optname,
                    unsigned long optval, int optlen, unsigned char is_val)
{
        int                     rv;
        pegnet_inprogress_t     op;
        unm_msg_t               msg;
        pegnet_queue_addr_t     remote = sock->qid;
        setsockopt_request_t    *req;
        setsockopt_response_t   *resp;

        DPRINTF(4, ("%s: sock:%p, level:%d, optname:%d, optval:%lx, "
                    "optlen:%d\n",
                    __FUNCTION__, sock, level, optname, optval, optlen));

        if (null_queue(remote)) {
                rv = -ENOTSOCK;
                goto done;
        }
        rv = host_begin_operation(sock, &op, op_setsockopt);  /* pre-condition
                                                                 the op-block*/
        if (rv != 0) {
                goto done;
        }
        req = (setsockopt_request_t *)&msg.body;
        req->hdr.opcode = UNM_H2P_OPCODE_SETSOCKOPT_REQUEST;
        req->hdr.op_id = op.op_id;
        req->hdr.ctx = sock->ctx.context;
        req->level = level;
        req->optname = optname;
        req->optval_ptr = optval;
	req->is_val = is_val;
        if (req->optval_ptr != 0) {
                req->optval_ptr |= PEGNET_HOST_ADDR;
        }
        req->optlen = optlen;

        UNM_MSGQ_HDR(msg.hdr, remote.major, remote.minor, HOST_REQUEST);

        rv = unm_send_h2p_msg(NX_GET_DEV_FROM_TNIC_ADAPTER(sock->adapter,sock->portnum),
			&msg, sizeof (setsockopt_request_t));
        if (rv == 0) {
                rv = host_end_operation(sock, &op);  /* post-conditioning */
                if (rv == 0) {
                        resp = (setsockopt_response_t *)&(op.msg.body);
                        rv = resp->ret_code;
                }
        } else {
                rv = -ENOMEM;
        }

  done:
        DPRINTF(5, ("%s: return(rv:%d); }\n", __FUNCTION__, rv));
        return (rv);
}  /* host_setsockopt */

/*
 * Called when a response is received for an setsockopt request.
 * Stores the result and wakes up the process that made the request.
 *
 * Parameters:
 *      msg     - The message received from the card.
 *      sid     - The sock context retreived for this response.
 *      trans   - The transaction of the corresponding request.
 *
 * Returns:
 *      Void.
 */
void unm_host_proc_setsockopt_response(nx_tnic_adapter_t *tnic_adapter,unm_msg_t *msg, 
			pegnet_sock_id_t *sid,pegnet_inprogress_t *trans)
{
        setsockopt_response_t   *resp = (setsockopt_response_t *)&msg->body;
        setsockopt_response_t   *req;

        DPRINTF(5, ("%s: Called\n", __FUNCTION__));

        req = (setsockopt_response_t *)&(trans->msg.body);
        req->ret_code = host_get_uint32_from_peg(resp->ret_code);
        wake_up(&trans->wq);
}

/*
 *
 */
int host_getsockopt(pegnet_sock_id_t *sock, int level, int optname,
                    uint64_t optval, int *optlen)
{
        int                     rv;
        pegnet_inprogress_t     op;
        unm_msg_t               msg;
        pegnet_queue_addr_t     remote = sock->qid;
        getsockopt_request_t    *req;
        getsockopt_response_t   *resp;

        DPRINTF(4, ("%s: sock:%p, level:%d, optname:%d, optval:%llx, "
                    "optlen:%llx): {\n",
                    __FUNCTION__, sock, level, optname, optval, optlen));

        if (null_queue(remote)) {
                rv = -ENOTSOCK;
                goto done;
        }

        rv = host_begin_operation(sock, &op, op_getsockopt);  /* pre-condition
                                                                 the op-block*/
        if (rv != 0) {
                goto done;
        }

        req = (getsockopt_request_t *)&msg.body;
        req->hdr.opcode = UNM_H2P_OPCODE_GETSOCKOPT_REQUEST;
        req->hdr.op_id = op.op_id;
        req->hdr.ctx = sock->ctx.context;
        req->level = level;
        req->optname = optname;
        req->optval_ptr = optval;
        if (req->optval_ptr != 0) {
                req->optval_ptr |= PEGNET_HOST_ADDR;
        }
        req->optlen = *optlen;

        UNM_MSGQ_HDR(msg.hdr, remote.major, remote.minor, HOST_REQUEST);

        rv = unm_send_h2p_msg(NX_GET_DEV_FROM_TNIC_ADAPTER(sock->adapter,sock->portnum),
			&msg, sizeof (getsockopt_request_t));
        if (rv == 0) {
                rv = host_end_operation(sock, &op);  /* post-conditioning */
                if (rv == 0) {
                        resp = (getsockopt_response_t *)&(op.msg.body);
                        rv = resp->ret_code;
                        *optlen = resp->optlen;
                }
        } else {
                rv = -ENOMEM;
        }

  done:
        DPRINTF(5, ("%s: return(rv:%d); }\n", __FUNCTION__, rv));
        return (rv);
}  /* host_getsockopt */

/*
 * Called when a response is received for an getsockopt request.
 * Stores the result and wakes up the process that made the request.
 *
 * Parameters:
 *      msg     - The message received from the card.
 *      sid     - The sock context retreived for this response.
 *      trans   - The transaction of the corresponding request.
 *
 * Returns:
 *      Void.
 */
void unm_host_proc_getsockopt_response(nx_tnic_adapter_t *tnic_adapter,unm_msg_t *msg, 
			pegnet_sock_id_t *sid,pegnet_inprogress_t *trans)
{
        getsockopt_response_t   *resp = (getsockopt_response_t *)&msg->body;
        getsockopt_response_t   *req;

        DPRINTF(5, ("%s: Called\n", __FUNCTION__));

        req = (getsockopt_response_t *)&(trans->msg.body);
        req->ret_code = host_get_uint32_from_peg(resp->ret_code);
        req->optlen = host_get_uint32_from_peg(resp->optlen);
        wake_up(&trans->wq);
}

/*
 *
 */
int host_ipt_setsockopt(pegnet_sock_id_t *sock, int level, int optname,
                        uint64_t optval, int optlen)
{
        int                             rv;
        pegnet_inprogress_t             op;
        unm_msg_t                       msg;
        pegnet_queue_addr_t             remote = sock->qid;
        iptsetsockopt_request_t         *req;
        iptsetsockopt_response_t        *resp;


        DPRINTF(4, ("%s: sock:%p, level:%d, optname:%d, optval:%lx, "
                    "optlen:%d\n",
                    __FUNCTION__, sock, level, optname, optval, optlen));

        if (null_queue(remote)) {
                rv = -ENOTSOCK;
                goto done;
        }

        /* pre-condition the op-block */
        rv = host_begin_operation(sock, &op, op_ipt_setsockopt);
        if (rv != 0) {
                goto done;
        }
        req = (iptsetsockopt_request_t *)&msg.body;
        req->hdr.opcode = UNM_H2P_OPCODE_IPTSETSOCKOPT_REQUEST;
        req->hdr.op_id = op.op_id;
        req->hdr.ctx = sock->ctx.context;
        req->level = level;
        req->optname = optname;
        req->optval_ptr = optval;
        if (req->optval_ptr != 0) {
                req->optval_ptr |= PEGNET_HOST_ADDR;
        }
        req->optlen = optlen;

        UNM_MSGQ_HDR(msg.hdr, remote.major, remote.minor, HOST_REQUEST);

        rv = unm_send_h2p_msg(NX_GET_DEV_FROM_TNIC_ADAPTER(sock->adapter,sock->portnum),
		&msg, sizeof (iptsetsockopt_request_t));
        if (rv == 0) {
                rv = host_end_operation(sock, &op);  /* post-conditioning */
                if (rv == 0) {
                        resp = (iptsetsockopt_response_t *)&op.msg.body;
                        rv = resp->ret_code;
                }
        } else {
                rv = -ENOMEM;
        }

  done:
        DPRINTF(5, ("%s: return(rv:%d); }\n", __FUNCTION__, rv));
        return (rv);
}  /* host_ipt_setsockopt */

/*
 * Called when a response is received for an ipt_setsockopt request.
 * Stores the result and wakes up the process that made the request.
 *
 * Parameters:
 *      msg     - The message received from the card.
 *      sid     - The sock context retreived for this response.
 *      trans   - The transaction of the corresponding request.
 *
 * Returns:
 *      Void.
 */
void unm_host_proc_iptsetsockopt_response(nx_tnic_adapter_t *tnic_adapter,unm_msg_t *msg,
                                pegnet_sock_id_t *sid,pegnet_inprogress_t *trans)
{
        iptsetsockopt_response_t        *resp;
        iptsetsockopt_response_t        *req;

        DPRINTF(5, ("%s: Called\n", __FUNCTION__));

        resp = (iptsetsockopt_response_t *)&msg->body;
        req = (iptsetsockopt_response_t *)&(trans->msg.body);
        req->ret_code = host_get_uint32_from_peg(resp->ret_code);
        wake_up(&trans->wq);
}

/*
 *
 */
int host_ipt_getsockopt(pegnet_sock_id_t *sock, int level, int optname,
                        uint64_t optval, int *optlen)
{
        int                             rv;
        pegnet_inprogress_t             op;
        unm_msg_t                       msg;
        pegnet_queue_addr_t             remote = sock->qid;
        iptgetsockopt_request_t         *req;
        iptgetsockopt_response_t        *resp;

        DPRINTF(4, ("%s: sock:%p, level:%d, optname:%d, optval:%llx, "
                    "optlen:%llx): {\n",
                    __FUNCTION__, sock, level, optname, optval, optlen));

        if (null_queue(remote)) {
                rv = -ENOTSOCK;
                goto done;
        }

        /* pre-condition the op-block */
        rv = host_begin_operation(sock, &op, op_ipt_getsockopt);
        if (rv != 0) {
                goto done;
        }

        req = (iptgetsockopt_request_t *)&msg.body;
        req->hdr.opcode = UNM_H2P_OPCODE_IPTGETSOCKOPT_REQUEST;
        req->hdr.op_id = op.op_id;
        req->hdr.ctx = sock->ctx.context;
        req->level = level;
        req->optname = optname;
        req->optval_ptr = optval;
        if (req->optval_ptr != 0) {
                req->optval_ptr |= PEGNET_HOST_ADDR;
        }
        req->optlen = *optlen;

        UNM_MSGQ_HDR(msg.hdr, remote.major, remote.minor, HOST_REQUEST);

        rv = unm_send_h2p_msg(NX_GET_DEV_FROM_TNIC_ADAPTER(sock->adapter,sock->portnum),
			&msg, sizeof (iptgetsockopt_request_t));
        if (rv == 0) {
                rv = host_end_operation(sock, &op);  /* post-conditioning */
                if (rv == 0) {
                        resp = (iptgetsockopt_response_t *)&op.msg.body;
                        rv = resp->ret_code;
                        *optlen = resp->optlen;
                }
        } else {
                rv = -ENOMEM;
        }

  done:
        DPRINTF(5, ("%s: return(rv:%d); }\n", __FUNCTION__, rv));
        return (rv);
}  /* host_ipt_getsockopt */

/*
 * Called when a response is received for an getsockopt request.
 * Stores the result and wakes up the process that made the request.
 *
 * Parameters:
 *      msg     - The message received from the card.
 *      sid     - The sock context retreived for this response.
 *      trans   - The transaction of the corresponding request.
 *
 * Returns:
 *      Void.
 */
void unm_host_proc_iptgetsockopt_response(nx_tnic_adapter_t *tnic_adapter,unm_msg_t *msg,
                             pegnet_sock_id_t *sid,pegnet_inprogress_t *trans)
{
        iptgetsockopt_response_t        *resp;
        iptgetsockopt_response_t        *req;

        DPRINTF(5, ("%s: Called\n", __FUNCTION__));

        resp = (iptgetsockopt_response_t *)&msg->body;
        req = (iptgetsockopt_response_t *)&(trans->msg.body);
        req->ret_code = host_get_uint32_from_peg(resp->ret_code);
        req->optlen = host_get_uint32_from_peg(resp->optlen);
        wake_up(&trans->wq);
}

/*
 *
 */
void nx_host_post_accept(pegnet_sock_id_t	*sid, unsigned int count)
{
        unm_msg_t               msg;
        UNM_MSGQ_HDR(msg.hdr,
			sid->free_accept_qid.major,
			sid->free_accept_qid.minor,
			UNKNOWN); /* Message type and body are
				      not important. This will
				      work like a ticket queue. */

#if 0
	printk("%s: sid %p, major = %x, minor = %x, count = %d\n",
			__FUNCTION__, sid,
			sid->accept.passive.free_accept_qid.major,
			sid->accept.passive.free_accept_qid.minor,
			count);
#endif

	for (;count--;) {
		if (unm_send_h2p_msg(NX_GET_DEV_FROM_TNIC_ADAPTER(sid->adapter,sid->portnum),
		    &msg, sizeof (msg))) {
			printk("%s: Could not send free accept to card\n",
					__FUNCTION__);
		}
	}

	return;
}

int host_listen(pegnet_sock_id_t* sock, int backlog, int do_bind)
{
        int                     rv;
        pegnet_inprogress_t     op;
        unm_msg_t		msg;
        pegnet_queue_addr_t     remote = sock->qid;
        listen_request_t        *req;
        listen_response_t       *resp;
	listener_group_t	*grp = sock->grp;

	DPRINTF(4, ("%s: sock:%p, backlog:%d\n", __FUNCTION__, sock, backlog));
	if(backlog==0){
		backlog=1;
	}
	if (null_queue(remote)) {
		rv = -ENOTSOCK;
		goto done;
	}

	
	rv = host_begin_operation(sock, &op, op_listen);  /* pre-condition the
							     op-block */
	if (rv != 0) {
		goto done;
	}

        req = (listen_request_t *)&msg.body;
        req->hdr.opcode = UNM_H2P_OPCODE_LISTEN_REQUEST;
        req->hdr.op_id = op.op_id;
        req->hdr.ctx = sock->ctx.context;
        req->backlog = backlog;
	req->lip_addr.v4 = grp->saddr.s_addr;
        req->lport = htons(grp->sport);
	req->reuseflg = grp->reuseflg;
	sock->reported_rcvbuf = grp->rcvbuf;
	req->rcvbuf = host_encode_uint32_for_peg(sock->reported_rcvbuf/2);
	req->listener_id = host_flowtbl_get_index(sock->adapter->p_ftable,sock->connID);
	req->do_bind = do_bind ? 1 : 0;
        req->grp_idx = sock->grp_idx;
        UNM_MSGQ_HDR(msg.hdr, remote.major, remote.minor, HOST_REQUEST);

        rv = unm_send_h2p_msg(NX_GET_DEV_FROM_TNIC_ADAPTER(sock->adapter,sock->portnum),
			&msg, sizeof (listen_request_t));
	if (rv == 0) {
		sock->listener = 1;
		rv = host_end_operation(sock, &op);  /* post-conditioning */
		if (rv == 0) {
                        resp = (listen_response_t *)&(op.msg.body);
                        rv = resp->ret_code;
                        sock->lport = htons(grp->sport);
                        sock->lip_addr=(ip_addr_t)(grp->saddr.s_addr);
			if(rv == 0) {
				nx_host_post_accept(sock, backlog);
			}	
	} else {
			sock->listener = 0;
	}
		} else {
		rv = -ENOMEM;
	}

  done:
	DPRINTF(5, ("%s: return(rv:%d); }\n", __FUNCTION__, rv));
	return rv;
}  /* host_listen */

/*
 * Called when a response is received for an listen request. Stores the result
 * and wakes up the process that made the request.
 *
 * Parameters:
 *      msg     - The message received from the card.
 *      sid     - The sock context retreived for this response.
 *      trans   - The transaction of the corresponding request.
 *
 * Returns:
 *      Void.
 */
void unm_host_proc_listen_response(nx_tnic_adapter_t *tnic_adapter,
				   unm_msg_t *msg, pegnet_sock_id_t *sid,
				   pegnet_inprogress_t *trans)
{
        listen_response_t       *resp = (listen_response_t *)&msg->body;
        listen_response_t       *req;
        DPRINTF(5, ("%s: Called\n", __FUNCTION__));

        req = (listen_response_t *)&(trans->msg.body);
        req->ret_code = host_get_uint32_from_peg(resp->ret_code);

	if (!req->ret_code) {
		host_assign_qid_from_peg(
				sid->free_accept_qid,
				resp->free_accept_qid);

#if 0
		printk("%s: Got major %x, minor %x\n",
				__FUNCTION__, sid->free_accept_qid.major,
				sid->free_accept_qid.minor);
#endif
	}

        wake_up(&trans->wq);
}


/*
 * Network Statistics requests
 */
int host_reset_stats(struct net_device *netdev, uint32_t port)
{
        reset_stats_request_t   *req;
        unm_msg_t               msg;

        req = (reset_stats_request_t *)&msg.body;
        memset(req, 0, sizeof (*req));
        req->hdr.opcode = UNM_H2P_OPCODE_RESET_STATS_REQUEST;
        req->port = host_encode_uint32_for_peg(port);

        UNM_MSGQ_HDR(msg.hdr, PEGNET_H2C_MAJOR, PEGNET_H2C_MINOR,
                     RESET_STATISTICS);

        return (unm_send_h2p_msg(netdev,&msg, sizeof (reset_stats_request_t)));
}  /* host_reset_stats */

int host_update_capture(struct net_device *netdev, uint32_t value)
{
	unm_msg_t                       msg;
	nx_set_capture_t		*req = (nx_set_capture_t *)&msg.body;
	int 				rv, if_id;

	req->hdr.opcode = NX_H2P_OPCODE_UPDATE_CAPTURE_REQUEST;
	req->value = host_encode_uint32_for_peg(value);
	if_id = nx_get_device_port(netdev);
	req->if_id = host_encode_uint32_for_peg((uint32_t)if_id);

        UNM_MSGQ_HDR(msg.hdr, PEGNET_H2C_MAJOR, PEGNET_H2C_MINOR,
                     HOST_REQUEST);

	if ((rv = unm_send_h2p_msg(netdev, &msg, sizeof(*req)))) {
		nx_print("unm_send_h2p_msg returned %d\n", rv);
	}

	return rv;
}

/*
 *
 */
int host_get_stats(nx_tnic_adapter_t *tnic, uint32_t port)
{
	struct net_device	*netdev;
        get_stats_request_t     *req;
	unm_msg_t       	msg;
	int			rv = 0;
        long			timeo;
	jiffies_t		now = jiffies;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	DEFINE_WAIT(wait);
#else
        DECLARE_WAITQUEUE(wait, current);
#endif

	netdev = NX_GET_DEV_FROM_TNIC_ADAPTER(tnic, port);
	if (netdev == NULL) {
		return (-EINVAL);
	}

	spin_lock_bh(&tnic->lock);

	if (!tnic->stats.pending && (now - tnic->stats.last_time) > (HZ / 2)) {
		req = (get_stats_request_t *)&msg.body;
		memset(req, 0, sizeof (*req));
		req->hdr.opcode = UNM_H2P_OPCODE_GET_STATS_REQUEST;
		req->ring_ctx = tnic->ring_ctx;
		req->dma_to_addr = nx_h2p_uint64(tnic->stats.phys);
		req->dma_size = nx_h2p_uint32(sizeof (netxen_pstats_t));

		UNM_MSGQ_HDR(msg.hdr, PEGNET_H2C_MAJOR, PEGNET_H2C_MINOR,
			     GET_STATISTICS);

		rv = unm_send_h2p_msg(netdev, &msg,
				      sizeof (get_stats_request_t));
		if (rv != 0) {
			nx_print("Error[%d]: Unable to send Statistics "
				 "request.\n", rv);
			goto done;
		}
		tnic->stats.pending = 1;
		tnic->stats.last_time = now;
	}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        prepare_to_wait(&tnic->stats.wq, &wait, TASK_INTERRUPTIBLE);
#else
        add_wait_queue(&tnic->stats.wq, &wait);
	set_current_state(TASK_INTERRUPTIBLE);
#endif
        timeo = HOST_TIMEOUT;
	rv = -ETIMEDOUT;

        while (timeo && tnic->stats.pending) {
		if (signal_pending(current)) {
                        rv = -EINTR;
                        goto finish_wait_queue;
                }
                spin_unlock_bh(&tnic->lock);
		timeo = schedule_timeout(timeo);
                spin_lock_bh(&tnic->lock);
        }

	if (!tnic->stats.pending) {
		rv = tnic->stats.ret_code;
	}

  finish_wait_queue:
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        finish_wait(&tnic->stats.wq, &wait);
#else
	set_current_state(TASK_RUNNING);
        remove_wait_queue(&tnic->stats.wq, &wait);
#endif

  done:
	tnic->stats.pending = 0;
	spin_unlock_bh(&tnic->lock);
	return (rv);
} /* host_getstats */

/*
 *
 */
void nx_proc_get_stats_reply(nx_tnic_adapter_t *tnic, unm_msg_t *msg,
			     pegnet_sock_id_t *sid, pegnet_inprogress_t *trans)
{
        get_stats_response_t    *resp = (get_stats_response_t *)&msg->body;

        spin_lock(&tnic->lock);

        if (!tnic->stats.pending) {
                nx_print("WARNING: No request pending.\n");
        }

	tnic->stats.ret_code = host_get_uint32_from_peg(resp->ret_code);
	tnic->stats.pending = 0;

        spin_unlock(&tnic->lock);

	wake_up_all(&tnic->stats.wq);
}

/*
 *
 */
int nx_lsa_get_stats(struct net_device *dev, uint8_t port,
		     struct net_device_stats *stats)
{
	nx_tnic_adapter_t	*tnic;
	int			rv;
	int			ii;
	netxen_pstats_t		*card_stats;

        tnic = nx_get_dev_tnic_adapter(dev);
	if (tnic == NULL) {
		nx_print("Adapter not found\n");
		return (-EINVAL);
	}

	if (port >= NX_TNIC_MAX_PORTS) {
		nx_print("Port[%u] not valid\n", port);
		return (-EINVAL);
	}

	rv = host_get_stats(tnic, port);
	if (rv) {
		nx_print("Getting stats from card failed\n");
		return (rv);
	}

	card_stats = tnic->stats.data;
	for (ii = 0; ii < NETXEN_NUM_PEGS; ++ii) {
		stats->rx_packets += card_stats->peg[ii].port[port].tcp_statistics.TcpInSegs;
		stats->tx_packets += card_stats->peg[ii].port[port].tcp_statistics.TcpOutSegs;
		stats->rx_bytes += card_stats->peg[ii].port[port].l2_statistics.L2RxBytes;
		stats->tx_bytes += card_stats->peg[ii].port[port].l2_statistics.L2TxBytes;
	}

	return (0);
}
