/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
/******************************************************************************
 * unm.h --
 *
 *
 *
 *
 *
 *
 *
 ******************************************************************************
 */


#ifndef __PSTACK_H__

#define __PSTACK_H__

#include <linux/version.h>
#include <linux/init.h> // Needed for the macros
#include <linux/notifier.h> // Needed for the macros
#include <linux/net.h>
#include <linux/ip.h>
#include <linux/tcp.h>
#include <linux/poll.h>
#include <linux/netfilter.h>
#include <linux/netfilter_arp.h>
#include <net/sock.h>
#include <linux/types.h>
#include <asm/byteorder.h>
#include "unm_host_pnic_types.h"
#include "netxen_linux_ioctl.h"

/*
 * Externals
 */
extern struct proto_ops host_inet_stream_ops;

/*
 * Offload rule node.
 */
typedef struct netxen_offload_rule_node_s
{
        struct list_head        links;
        netxen_offload_rule_t   rule;
} netxen_offload_rule_node_t;

/*
 * Offload rules list.
 */
typedef struct netxen_offload_rule_list_s
{
        struct list_head        list;
        __uint32_t              cnt;
        rwlock_t                rwlock;
} netxen_offload_rule_list_t;

/*
 * Function Prototypes
 */
int unm_inet_release(struct socket *sock);
int unm_inet_bind(struct socket *sock, struct sockaddr *uaddr, int addr_len);

int unm_inet_connect(struct socket *sock, struct sockaddr * uaddr,
                     int addr_len, int flags);
int unm_inet_recvmsg(struct kiocb *iocb, struct socket *sock,
		     struct msghdr *msg, size_t size, int flags);

int unm_sock_no_socketpair(struct socket *sock1, struct socket *sock2);
int unm_inet_accept(struct socket *sock, struct socket *newsock, int flags);

int unm_inet_getname(struct socket *sock, struct sockaddr *uaddr,
                     int *uaddr_len, int peer);
unsigned int unm_tcp_poll(struct file *file, struct socket *sock, 
                          poll_table *wait);

int unm_inet_ioctl(struct socket *sock, unsigned int cmd, unsigned long arg);
int unm_inet_listen(struct socket *sock, int backlog);
int unm_inet_shutdown(struct socket *sock, int how);
int unm_inet_setsockopt(struct socket *sock, int level, int optname,
                        char *optval, int optlen);
int unm_inet_getsockopt(struct socket *sock, int level, int optname,
                        char *optval, int *optlen);
int unm_inet_sendmsg(struct kiocb *iocb, struct socket *sock,
		     struct msghdr *msg, size_t size);
ssize_t unm_tcp_sendpage(struct socket *sock, struct page *page, int offset, 
                         size_t size, int flags);
int unm_copy_sockopts(struct socket *sock);
void unm_register_netdev_notifiers(void);
void unm_unregister_netdev_notifiers(void);

int unm_inet_dgrams_ioctl(struct socket *sock, unsigned int cmd,
			  unsigned long arg);
#if (defined (CONFIG_COMPAT)) && (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,17))
int unm_compat_sock_common_setsockopt(struct socket *sock, int level, int optname,
                                  char  *optval, int optlen);
int unm_compat_sock_common_getsockopt(struct socket *sock, int level, int optname,
                                  char *optval, int *optlen);
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
extern struct sock *__tcp_v4_lookup(u32 saddr, u16 sport, u32 daddr,
				    u16 dport, int dif);
#endif /* LINUX_VERSION */

#endif // __UNM_H__

