/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
/*
 * Main module
 */

#include <linux/module.h>
#include <linux/kernel.h> // Needed for KERN_ALERT
#include <linux/version.h>
#include <asm/pgtable.h>  // For PAGE_KERNEL_EXEC
#include "unm_host_pnic_types.h"
#include "pstack.h"
#include "nx_intercept.h"


#ifdef  UNM_CONFIG_COMPAT
#include <linux/ioctl32.h>
#define UNM_CONFIG_COMPAT
#endif

struct proto_ops		host_inet_stream_ops;

nx_intercept_registry_t		nx_intercept_registry = {
lsa_stream_ops: {
	release: unm_inet_release,
	 socketpair: unm_sock_no_socketpair,
	 accept: unm_inet_accept,
	 getname: unm_inet_getname,
	 poll: unm_tcp_poll,
	 ioctl: unm_inet_ioctl,
	 shutdown: unm_inet_shutdown,
	 sendmsg: unm_inet_sendmsg,
	 recvmsg: unm_inet_recvmsg,
	 setsockopt: unm_inet_setsockopt,
	 getsockopt: unm_inet_getsockopt,
	 sendpage: unm_tcp_sendpage,
	 listen: unm_inet_listen,
	 connect: unm_inet_connect,
#if (defined (CONFIG_COMPAT)) && (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,17))
        compat_setsockopt: unm_compat_sock_common_setsockopt,
        compat_getsockopt: unm_compat_sock_common_getsockopt,
#endif
	 
		},
lsa_dgram_ioctl: unm_inet_dgrams_ioctl,
};

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
kmem_cache_t *nx_sk_cachep;
#endif

nx_ioctl_t native_inet_dgrams_ioctl;

nx_hbucket_head_t g_adapter_head;
atomic_t        g_devid;
index_table_t	g_listener_grp_tbl;

int pstack_init(void)
{
	nx_intercept_info_t	intercept_info;

        printk(KERN_ALERT "pstack_module_init: \n");

	if (nx_lsa_register_ops(&nx_intercept_registry, &intercept_info)) {
		return -EIO;
	}

	memcpy(&host_inet_stream_ops, &intercept_info.host_stream_ops,
	       sizeof (host_inet_stream_ops));
        native_inet_dgrams_ioctl = intercept_info.host_dgram_ioctl;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
        nx_sk_cachep = kmem_cache_create("nx_sock", sizeof(struct tcp_sock), 0,
                                         SLAB_HWCACHE_ALIGN, 0, 0);
        if (!nx_sk_cachep) {
            printk(KERN_ERR "%s: Can't create protocol sock SLAB cache!\n",
                    __FUNCTION__);
        }
#endif

#ifdef  UNM_CONFIG_COMPAT
        /* Our ioctls are all "32/64bit compatible" */
        if (register_ioctl32_conversion(SIOC_NETXEN_OFFLOAD, NULL)) {
                printk(KERN_ERR "%s: Error registering ioctl32 translation.\n",
                       __FUNCTION__);
        }
#endif  /* UNM_CONFIG_COMPAT */
	INIT_NX_HBUCKET_HEAD(&g_adapter_head);
	/*
	 * Init route table and pruning task.
	 */
	nx_route_tbl_init();
        nx_init_pruning_task();
	
	atomic_set(&g_devid,0);
	host_indextbl_init(&g_listener_grp_tbl);
        unm_register_netdev_notifiers();
        return 0;
}

extern void unm_unregister_netdev_notifiers(void);

void pstack_exit(void)
{
        printk(KERN_ALERT "pstack_module_exit:\n");

#ifdef UNM_CONFIG_COMPAT
        if (unregister_ioctl32_conversion(SIOC_NETXEN_OFFLOAD)) {
                printk(KERN_ERR "%s: Error unregistering ioctl32 "
                       "translation.\n", __FUNCTION__);
        }
#endif  /* UNM_CONFIG_COMPAT */

        unm_unregister_netdev_notifiers();
	host_indextbl_destroy(&g_listener_grp_tbl);

	nx_lsa_unregister_ops();

	nx_route_tbl_destroy();
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
	kmem_cache_destroy(nx_sk_cachep);
#endif 
}

